"""@package ecohydrolib.metadata
    
@brief Classes for writing and reading metadata for Ecohydrology workflows

This software is provided free of charge under the New BSD License. Please see
the following license information:

Copyright (c) 2013, University of North Carolina at Chapel Hill
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of the University of North Carolina at Chapel Hill nor the
      names of its contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE UNIVERSITY OF NORTH CAROLINA AT CHAPEL HILL
BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


@author Brian Miles <brian_miles@unc.edu>

@todo Refactor storage of climate station to embedded database
"""
import os
import time
import errno
import ConfigParser
from datetime import datetime

import ecohydrolib
import ecohydrolib.util

class MetadataEntity(object):
    
    FMT_DATE = '%Y-%m-%d %H:%M:%S'
    
    """ Abstract class for encoding structured data to be written to a metadata store
    """
    def writeToMetadata(self, context):
        """ Write structured entity to metadata store for a given project directory
        
            @param context Context object containing projectDir, the path of the project whose 
            metadata store is to be written to
        """
        pass
    
    @classmethod
    def readFromMetadata(cls, context, fqId):
        """ Read structured entity from metadata store for a given project directory
        
            @param context Context object containing projectDir, the path of the project whose 
            metadata store is to be read from
            @param fqId String representing the fully qualified ID of the entity
            
            Implementations should return an instance of themselves containing data read from the metadata store
            
            @raise KeyError if entity is not in metadata
        """
        pass


class ClimatePointStation(MetadataEntity):
    
    VAR_PRECIP = 'prcp'
    VAR_SNOW = 'snow'
    VAR_TMIN = 'tmin'
    VAR_TMAX = 'tmax'
    VAR_RH = 'rh'
    VAR_PAR = 'par'
    VAR_WIND2M = 'wspd2m'
    
    def __init__(self):
        self.type = None
        self.id = None
        self.longitude = None
        self.latitude = None
        self.elevation = None
        self.name = None
        self.startDate = None
        self.endDate = None
        self.variables = []
        self.data = None
        self.variablesData = {}
        
    def writeToMetadata(self, context):
        """ Write ClimatePointStation data to climate point section of metadata for
            a given project directory
        
            @param context Context object containing projectDir, the path of the project whose 
            metadata store is to be written to
        """
        fqId = self.type + GenericMetadata.COMPOUND_KEY_SEP + self.id
        fqId = fqId.lower()

        climatePoints = GenericMetadata.readClimatePointEntries(context)
        try:
            stations = climatePoints['stations'].split(GenericMetadata.VALUE_DELIM)
        except KeyError:
            stations = []
        # Write station metadata (overwrite if already present)
        keys = []
        values = []
        if fqId not in stations:
            stations.append(fqId)
            stationsStr = GenericMetadata.VALUE_DELIM.join(stations)
            keys.append('stations'); values.append(stationsStr)
        # Write attributes for station
        keyProto = 'station' + GenericMetadata.COMPOUND_KEY_SEP + fqId + GenericMetadata.COMPOUND_KEY_SEP 
        longitude = keyProto + 'longitude'
        keys.append(longitude); values.append(self.longitude)
        latitude = keyProto + 'latitude'
        keys.append(latitude); values.append(self.latitude)
        elevation = keyProto + 'elevation'
        keys.append(elevation); values.append(self.elevation)
        name = keyProto + 'name'
        keys.append(name); values.append(self.name)
        if self.startDate:
            startDate = keyProto + 'startdate'
            keys.append(startDate); values.append(self.startDate.strftime(ClimatePointStation.FMT_DATE))
        if self.endDate:
            endDate = keyProto + 'enddate'
            keys.append(endDate); values.append(self.endDate.strftime(ClimatePointStation.FMT_DATE))
        if self.variables:
            variablesKey = keyProto + 'variables'
            variablesValue = GenericMetadata.VALUE_DELIM.join(self.variables)
            keys.append(variablesKey); values.append(variablesValue)
        if self.data != None:
            data = keyProto + 'data'
            keys.append(data); values.append(self.data)
        elif self.variablesData:
            # Try to write data entries for each variable separately
            vars = self.variablesData.keys()
            for var in vars:
                varKey = keyProto + var + GenericMetadata.COMPOUND_KEY_SEP + 'data'
                keys.append(varKey); values.append(self.variablesData[var])
        GenericMetadata.writeClimatePointEntries(context, keys, values)
    
    @classmethod
    def readFromMetadata(cls, context, fqId):
        """ Read ClimatePointStation data from climate point section of metadata for
            a given project directory
        
            @param context Context object containing projectDir, the path of the project whose 
            metadata store is to be read from
            @param fqId String representing the fully qualified station ID: <type>_<id>
            
            @return A new ClimatePointStation instance with data populated from metadata
            
            @raise KeyError if required field is not in metadata
        """
        newInstance = ClimatePointStation()
        (newInstance.type, newInstance.id) = fqId.split(GenericMetadata.COMPOUND_KEY_SEP)

        climate = GenericMetadata.readClimatePointEntries(context)
        
        keyProto = 'station' + GenericMetadata.COMPOUND_KEY_SEP + fqId + GenericMetadata.COMPOUND_KEY_SEP
        longitude = keyProto + 'longitude'
        newInstance.longitude = float(climate[longitude])
        latitude = keyProto + 'latitude'
        newInstance.latitude = float(climate[latitude])
        elevation = keyProto + 'elevation'
        newInstance.elevation = float(climate[elevation])
        name = keyProto + 'name'
        newInstance.name = climate[name]      
        startDate = keyProto + 'startdate'
        try:
            newInstance.startDate = datetime.strptime(climate[startDate], ClimatePointStation.FMT_DATE)
        except KeyError:
            pass    
        endDate = keyProto + 'enddate'
        try:
            newInstance.endDate = datetime.strptime(climate[endDate], ClimatePointStation.FMT_DATE)
        except KeyError:
            pass
        variablesKey = keyProto + 'variables'
        try:
            newInstance.variables = climate[variablesKey].split(GenericMetadata.VALUE_DELIM)
        except KeyError:
            pass
        data = keyProto + 'data'
        try:
            newInstance.data = climate[data]
        except KeyError:
            pass
        try:
            for var in newInstance.variables:
                varKey = keyProto + var + GenericMetadata.COMPOUND_KEY_SEP + 'data'
                newInstance.variablesData[var] = climate[varKey]
        except KeyError:
            pass
       
        return newInstance


class ModelRun(MetadataEntity):

    def __init__(self, modelType=None):
        self.modelType = modelType
        self.runNumber = None
        self.description = None
        self.date = None
        self.command = None
        self.output = None
        
    def writeToMetadata(self, context):
        """ Write ModelRun data to model run section of metadata for
            a given project directory
            
            @note Will set run number to value to be stored in metadata
        
            @param context Context object containing projectDir, the path of the project whose 
            metadata store is to be written to
            @exception Exception if model type has not known
            @exception Exception if section is not a valid GenericMetadata section
        """
        if self.modelType not in GenericMetadata.MODEL_TYPES:
            raise Exception("Model type %s is not among known model types: %s" % (self.modelType, str(GenericMetadata.MODEL_TYPES) ) )
        
        modelRunEntries = GenericMetadata.readModelRunEntries(context)
        try:
            runs = modelRunEntries['runs'].split(GenericMetadata.VALUE_DELIM)
        except KeyError:
            runs = []
        
        # Collected model entry and keys and values into lists so we can write to metadata store in batch
        keys = []
        values = []
            
        # Generate unique identifier for this model run.  Unique ID is a combination of model type and a number
        entryNumber = 1
        fqId = self.modelType + GenericMetadata.KEY_SEP + str(entryNumber)
        while fqId in runs:
            entryNumber += 1
            fqId = self.modelType + GenericMetadata.KEY_SEP + str(entryNumber)
        self.runNumber = entryNumber
        # Add new run to list of runs
        runs.append(fqId)
        runsStr = GenericMetadata.VALUE_DELIM.join(runs)
        keys.append('runs'); values.append(runsStr)
        # Write attributes for run
        keyProto = fqId + GenericMetadata.KEY_SEP
        runDate = keyProto + 'date_utc'
        keys.append(runDate); values.append( self.date.strftime(ModelRun.FMT_DATE) )
        runDesc = keyProto + 'description'
        keys.append(runDesc); values.append(self.description)
        runCmd = keyProto + 'command'
        keys.append(runCmd); values.append(self.command)
        runOutput = keyProto + 'output'
        keys.append(runOutput); values.append(self.output)
        # Write to metadata
        GenericMetadata.writeModelRunEntries(context, keys, values)
        
    @classmethod
    def readFromMetadata(cls, context, fqId):
        """ Read ModelRun data from model run section of metadata for
            a given project directory
        
            @param context Context object containing projectDir, the path of the project whose 
            metadata store is to be read from
            @param fqId String representing the fully qualified ID of the model run: <model_type>_<run_number>
            
            @return A new ModelRun instance with data populated from metadata
            
            @raise KeyError if required field is not in metadata
        """      
        newInstance = ModelRun()
        (newInstance.modelType, newInstance.runNumber) = fqId.split(GenericMetadata.KEY_SEP)
        
        modelRunEntries = GenericMetadata.readModelRunEntries(context)
        keyProto = fqId + GenericMetadata.KEY_SEP
        
        runDate = keyProto + 'date_utc'
        newInstance.date = datetime.strptime(modelRunEntries[runDate], ModelRun.FMT_DATE)
        runDesc = keyProto + 'description'
        newInstance.description = modelRunEntries[runDesc]
        runCmd = keyProto + 'command'
        newInstance.command = modelRunEntries[runCmd]
        runOutput = keyProto + 'output'
        newInstance.output = modelRunEntries[runOutput]
        
        return newInstance
    

class AssetProvenance(MetadataEntity):
    
    def __init__(self, section=None):
        self.section = section
        self.name = None
        self.dcIdentifier = None
        self.dcSource = None
        self.dcTitle = None
        self.dcDate = datetime.now()
        self.dcPublisher = None
        self.dcDescription = None
        self.processingNotes = None
        
    def writeToMetadata(self, context):
        """ Write AssetProvenance data to provenance section of metadata for
            a given project directory
        
            @param context Context object containing projectDir, the path of the project whose 
            metadata store is to be written to
            @exception Exception if section is not a valid GenericMetadata section
        """
        fqId = self.section + GenericMetadata.COMPOUND_KEY_SEP + self.name
        fqId = fqId.lower()
        
        # Write self to the appropriate section
        GenericMetadata.writeEntryToSection(context, self.section, self.name, self.dcIdentifier)
        
        # Write to provenance section
        provenanceEntries = GenericMetadata.readProvenanceEntries(context)
        try:
            entities = provenanceEntries['entities'].split(GenericMetadata.VALUE_DELIM)
        except KeyError:
            entities = []
        # Write entity metadata (overwrite if already present)
        keys = []
        values = []
        if fqId not in entities:
            entities.append(fqId)
            entitiesStr = GenericMetadata.VALUE_DELIM.join(entities)
            keys.append('entities'); values.append(entitiesStr)
        # Write attributes for entity
        keyProto = fqId + GenericMetadata.COMPOUND_KEY_SEP
        dcIdentifier = keyProto + 'dc.identifier'
        keys.append(dcIdentifier); values.append(self.dcIdentifier)
        dcSource = keyProto + 'dc.source'
        keys.append(dcSource); values.append(self.dcSource)
        dcTitle = keyProto + 'dc.title'
        keys.append(dcTitle); values.append(self.dcTitle)
        if self.dcDate:
            dcDate = keyProto + 'dc.date'
            keys.append(dcDate); values.append(self.dcDate.strftime(AssetProvenance.FMT_DATE))
        dcPublisher = keyProto + 'dc.publisher'
        keys.append(dcPublisher); values.append(self.dcPublisher)
        dcDescription = keyProto + 'dc.description'
        keys.append(dcDescription); values.append(self.dcDescription)
        processingNotes = keyProto + 'processing_notes'
        keys.append(processingNotes); values.append(self.processingNotes)
        GenericMetadata.writeProvenanceEntries(context, keys, values)
    
    @classmethod
    def readFromMetadata(cls, context, fqId):
        """ Read AssetProvenance data from provenance section of metadata for
            a given project directory
        
            @param context Context object containing projectDir, the path of the project whose 
            metadata store is to be read from
            @param fqId String representing the fully qualified ID of the asset: <section>_<name>
            
            @return A new AssetProvenance instance with data populated from metadata
            
            @raise KeyError if required field is not in metadata
        """
        newInstance = AssetProvenance()
        (newInstance.section, newInstance.name) = fqId.split(GenericMetadata.COMPOUND_KEY_SEP)
        
        provenance = GenericMetadata.readProvenanceEntries(context)
        keyProto = fqId + GenericMetadata.COMPOUND_KEY_SEP
        dcIdentifier = keyProto + 'dc.identifier'
        newInstance.dcIdentifier = provenance[dcIdentifier]
        dcSource = keyProto + 'dc.source'
        newInstance.dcSource = provenance[dcSource]
        dcTitle = keyProto + 'dc.title'
        newInstance.dcTitle = provenance[dcTitle]
        dcDate = keyProto + 'dc.date'
        newInstance.dcDate = datetime.strptime(provenance[dcDate], AssetProvenance.FMT_DATE)
        dcPublisher = keyProto + 'dc.publisher'
        newInstance.dcPublisher = provenance[dcPublisher]
        dcDescription = keyProto + 'dc.description'
        newInstance.dcDescription = provenance[dcDescription]
        processingNotes = keyProto + 'processing_notes'
        newInstance.processingNotes = provenance[processingNotes]
        
        return newInstance
    

class MetadataVersionError(ConfigParser.Error):
    def __init__(self, metadataVersion):
        self.metadataVersion = metadataVersion
        self._ecohydrolibVersion = ecohydrolib.__version__
    def __str__(self):
        return repr("Version %s of EcohydroLib was used to write the metadata, but you are running version %s of EcohydroLib" %\
                    (self.metadataVersion, self._ecohydrolibVersion) )    

class GenericMetadata(object):
    """ Handles metadata persistance.
    
        @note All keys are stored in lower case.
        @note This object is stateless, all methods are static, writes to metadata store
        are written immediately.
        
        @todo Implement lock file semantics as decorators
    """
    _ecohydrolibVersion = ecohydrolib.__version__

    VALUE_DELIM = ','
    KEY_SEP = '_'
    COMPOUND_KEY_SEP = '/' # Used for keys that may contain KEY_SEP
    METADATA_FILENAME = 'metadata.txt'
    METADATA_LOCKFILE = 'metadata.txt.lock'
    VERSION_KEY = 'ecohydrolib_version'
    
    ECOHYDROLIB_SECION = 'ecohydrolib'
    MANIFEST_SECTION = 'manifest'
    PROVENANCE_SECTION = 'provenance'
    HISTORY_SECTION = 'history'
    STUDY_AREA_SECTION = 'study_area'
    GRASS_SECTION = 'grass'
    CLIMATE_POINT_SECTION = 'climate_point'
    CLIMATE_GRID_SECTION = 'climate_grid'
    MODEL_RUN_SECTION = 'model_run'
    SECTIONS = [ECOHYDROLIB_SECION, MANIFEST_SECTION, PROVENANCE_SECTION, HISTORY_SECTION,
                STUDY_AREA_SECTION, GRASS_SECTION, CLIMATE_POINT_SECTION, CLIMATE_GRID_SECTION,
                MODEL_RUN_SECTION]
    
    HISTORY_PROTO = "processing%sstep%s" % (KEY_SEP, KEY_SEP)

    # Raster type list (this should probably be a dict)
    RASTER_TYPE_LC = 'landcover'
    RASTER_TYPE_ROOF = 'roof_connectivity'
    RASTER_TYPE_SOIL = 'soil'
    RASTER_TYPE_LAI = 'lai'
    RASTER_TYPE_PATCH = 'patch'
    RASTER_TYPE_ZONE = 'zone'
    RASTER_TYPE_ISOHYET = 'isohyet'
    RASTER_TYPE_LEAFC = 'leafc'
    RASTER_TYPE_ROOTDEPTH = 'rootdepth'
    RASTER_TYPE_ROAD = 'roads'
    RASTER_TYPE_STREAM_BURNED_DEM = 'stream_burned_dem'
    RASTER_TYPES = [RASTER_TYPE_LC, RASTER_TYPE_ROOF, RASTER_TYPE_SOIL, RASTER_TYPE_LAI, 
                    RASTER_TYPE_PATCH, RASTER_TYPE_ZONE, RASTER_TYPE_ISOHYET,
                    RASTER_TYPE_LEAFC, RASTER_TYPE_ROOTDEPTH, RASTER_TYPE_ROAD,
                    RASTER_TYPE_STREAM_BURNED_DEM]

    # Model type list
    MODEL_TYPES = []

    @staticmethod
    def getCommandLine():
        """ Return string representing original command line, as close as possible,
            used to run the command.  Will convert all paths in the command line to
            absolute path, if a non-path element has spaces in it, they will be
            quoted.
            
            @return String with each element of sys.argv separated by a space
        """
        import sys, os
        cmdline = os.path.abspath(sys.argv[0])
        for elem in sys.argv[1:]:
            cmdline += ' ' + ecohydrolib.util.getAbsolutePathOfItem(elem)
        return cmdline 


    @staticmethod
    def checkMetadataVersion(projectDir):
        """ Check if metadata store is compatible with current version of ecohydrolib. Accepts
            project directory as this method is used in the constructor to the Context class.
        
            @param projectDir, the path of the project whose metadata store is to be written to
            @raise MetadataVersionError if a version already exists in the metadata store
            and is different than GenericMetadata._ecohydrolibVersion
        """
        metadataFilepath = os.path.join(projectDir, GenericMetadata.METADATA_FILENAME)
        if os.path.exists(metadataFilepath):
            if not os.access(metadataFilepath, os.R_OK):
                raise IOError(errno.EACCES, "Unable to read metadata store for project %s" % \
                              (projectDir,))
            # Read metadata store
            config = ConfigParser.RawConfigParser()
            config.read(metadataFilepath)
            if config.has_section(GenericMetadata.ECOHYDROLIB_SECION):
                if config.has_option(GenericMetadata.ECOHYDROLIB_SECION, \
                                 GenericMetadata.VERSION_KEY):
                    metadataVersion = config.get(GenericMetadata.ECOHYDROLIB_SECION, \
                                     GenericMetadata.VERSION_KEY)
                    if metadataVersion != GenericMetadata._ecohydrolibVersion:
                        raise MetadataVersionError(metadataVersion)


    @staticmethod
    def _writeVersionToMetadata(config):
        """ Write EcohydroLib version to ECOHYDROLIB_SECION of metadata.
        
            @param config ConfigParser to write version information to
            @raise MetadataVersionError if a version already exists in the metadata store
            and is different than GenericMetadata._ecohydrolibVersion
        """
        if not config.has_section(GenericMetadata.ECOHYDROLIB_SECION):
            config.add_section(GenericMetadata.ECOHYDROLIB_SECION)
        
        if not config.has_option(GenericMetadata.ECOHYDROLIB_SECION, \
                             GenericMetadata.VERSION_KEY):
            config.set(GenericMetadata.ECOHYDROLIB_SECION, \
                       GenericMetadata.VERSION_KEY, GenericMetadata._ecohydrolibVersion)
            return
            
        metadataVersion = config.get(GenericMetadata.ECOHYDROLIB_SECION, \
                                     GenericMetadata.VERSION_KEY)
        if metadataVersion != GenericMetadata._ecohydrolibVersion:
            raise MetadataVersionError(metadataVersion)
        
    
    @staticmethod
    def deleteEntryFromSection(context, section, key, callback=None):
        """ Delete an entry from the given section of the metadata store for a given project.
        
            @param context Context object containing projectDir, the path of the project whose 
            metadata store is to be deleted from
            @param section The section the key is to be deleted from
            @param key The key to be deleted from the given section of the project metadata
            @param callback A function that should be called before deleting the entry.  The function takes as
            input the config object.
        
            @raise IOError(errno.EACCES) if the metadata store for the project is not writable
            @raise Exception if section is not a valid GenericMetadata section
            @raise MetadataVersionError if existing version information in metadata store
            does not match version of currently running EcohydroLib.
        """
        projectDir = context.projectDir
        if section not in GenericMetadata.SECTIONS:
            raise Exception( "%s is an unknown section" % (section,) )
        lockFilepath = os.path.join(projectDir, GenericMetadata.METADATA_LOCKFILE)
        metadataFilepath = os.path.join(projectDir, GenericMetadata.METADATA_FILENAME)
        if os.path.exists(metadataFilepath):
            if not os.access(metadataFilepath, os.W_OK):
                raise IOError(errno.EACCES, "Unable to write to metadata store for project %s" % \
                              (projectDir,))
        else:
            if not os.access(projectDir, os.W_OK):
                raise IOError(errno.EACCES, "Unable to write to metadata store for project %s" % \
                              (projectDir,))
            # Create metadata file as it does not exist
            metadataFD = open(metadataFilepath, 'w')
            metadataFD.close()
        
        # Wait for lockfile to be relinquished
        while os.path.exists(lockFilepath):
            time.sleep(5)
        # Write lock file
        open(lockFilepath, 'w').close()
        
        # Read metadata store
        config = ConfigParser.RawConfigParser()
        config.read(metadataFilepath)
        GenericMetadata._writeVersionToMetadata(config)
        
        if callback:
            callback(config)
        
        # Delete entry
        if config.has_section(section):
            config.remove_option(section, key)
            # Write metadata store
            config.write(open(metadataFilepath, 'w'))
        
        # Remove lock file
        os.unlink(lockFilepath)
    
    
    @staticmethod
    def writeEntryToSection(context, section, key, value, callback=None):
        """ Write an entry in the given section to the metadata store for a given project. 
        
            @note Will overwrite the value for a key that already exists
            
            @param context Context object containing projectDir, the path of the project whose 
            metadata store is to be written to
            @param section The section the key is to be written to
            @param key The key to be written to the given section of the project metadata
            @param value The value to be written for key stored in the given section of the project metadata
            @param callback A function that should be called before writing the entry.  The function takes as
            input the config object.
            
            @raise IOError(errno.EACCES) if the metadata store for the project is not writable
            @raise Exception if section is not a valid GenericMetadata section
            @raise MetadataVersionError if existing version information in metadata store
            does not match version of currently running EcohydroLib.
        """
        projectDir = context.projectDir
        if section not in GenericMetadata.SECTIONS:
            raise Exception( "%s is an unknown section" % (section,) )
        lockFilepath = os.path.join(projectDir, GenericMetadata.METADATA_LOCKFILE)
        metadataFilepath = os.path.join(projectDir, GenericMetadata.METADATA_FILENAME)
        if os.path.exists(metadataFilepath):
            if not os.access(metadataFilepath, os.W_OK):
                raise IOError(errno.EACCES, "Unable to write to metadata store for project %s" % \
                              (projectDir,))
        else:
            if not os.access(projectDir, os.W_OK):
                raise IOError(errno.EACCES, "Unable to write to metadata store for project %s" % \
                              (projectDir,))
            # Create metadata file as it does not exist
            metadataFD = open(metadataFilepath, 'w')
            metadataFD.close()
        
        # Wait for lockfile to be relinquished
        while os.path.exists(lockFilepath):
            time.sleep(5)
        # Write lock file
        open(lockFilepath, 'w').close()
        
        # Read metadata store
        config = ConfigParser.RawConfigParser()
        config.read(metadataFilepath)
        GenericMetadata._writeVersionToMetadata(config)
        
        if callback:
            callback(config)
        
        # Write new entry
        if not config.has_section(section):
            config.add_section(section)
        config.set(section, key, value)
        # Write metadata store
        config.write(open(metadataFilepath, 'w'))
        
        # Remove lock file
        os.unlink(lockFilepath)
        
        
    @staticmethod
    def _writeEntriesToSection(projectDir, section, keys, values, callback=None):
        """ Write entries in the given section to the metadata store for a given project. 
        
            @note Will overwrite the value for each key that already exists
            
            @param projectDir Path of the project whose metadata store is to be written to
            @param section The section the keys are to be written to
            @param keys List of keys to be written to the given section of the project metadata
            @param values List of values to be written for key stored in the given section of the project metadata
            @param callback A function that should be called before writing the entry.  The function takes as
            input the config object.
            
            @raise IOError(errno.EACCES) if the metadata store for the project is not writable
            @raise Exception if len(keys) != len(values)
            @raise MetadataVersionError if existing version information in metadata store
            does not match version of currently running EcohydroLib.
        """
        numKeys = len(keys)
        if numKeys != len(values):
            raise Exception( "%d keys specified for %d values" % (numKeys, len(values)) )
        
        lockFilepath = os.path.join(projectDir, GenericMetadata.METADATA_LOCKFILE)
        metadataFilepath = os.path.join(projectDir, GenericMetadata.METADATA_FILENAME)
        if os.path.exists(metadataFilepath):
            if not os.access(metadataFilepath, os.W_OK):
                raise IOError(errno.EACCES, "Unable to write to metadata store for project %s" % \
                              (projectDir,))
        else:
            if not os.access(projectDir, os.W_OK):
                raise IOError(errno.EACCES, "Unable to write to metadata store for project %s" % \
                              (projectDir,))
            # Create metadata file as it does not exist
            metadataFD = open(metadataFilepath, 'w')
            metadataFD.close()
        
        # Wait for lockfile to be relinquished
        while os.path.exists(lockFilepath):
            time.sleep(5)
        # Write lock file
        open(lockFilepath, 'w').close()
        
        # Read metadata store
        config = ConfigParser.RawConfigParser()
        config.read(metadataFilepath)
        GenericMetadata._writeVersionToMetadata(config)
        
        if callback:
            callback(config)
        
        # Write new entries
        if not config.has_section(section):
            config.add_section(section)
        for i in xrange(numKeys):
            config.set(section, keys[i], values[i])
        # Write metadata store
        config.write(open(metadataFilepath, 'w'))
        
        # Remove lock file
        os.unlink(lockFilepath)
    
    
    @staticmethod
    def writeManifestEntry(context, key, value):
        """ Write a manifest entry to the metadata store for a given project.
            
            @note Will overwrite the value for a key that already exists
        
            @param context Context object containing projectDir, the path of the project whose 
            metadata store is to be written to
            @param key The key to be written to the manifest section of the project metadata
            @param value The value to be written for key stored in the manifest section of the project metadata
            
            @exception IOError(errno.EACCES) if the metadata store for the project is not writable
        """
        GenericMetadata.writeEntryToSection(context, GenericMetadata.MANIFEST_SECTION, key, value)
     
        
    @staticmethod 
    def writeStudyAreaEntry(context, key, value):
        """ Write a study area entry to the metadata store for a given project.
            
            @note Will overwrite the value for a key that already exists
        
            @param context Context object containing projectDir, the path of the project whose 
            metadata store is to be written to
            @param key The key to be written to the study area section of the project metadata
            @param value The value to be written for key stored in the study area section of the project metadata
            
            @exception IOError(errno.EACCES) if the metadata store for the project is not writable
        """
        GenericMetadata.writeEntryToSection(context, GenericMetadata.STUDY_AREA_SECTION, key, value)
    
    
    @staticmethod 
    def writeGRASSEntry(context, key, value):
        """ Write a GRASS entry to the metadata store for a given project.  Typically used to 
            record the location within a project directory of a GRASS location and mapset, as well
            as to note the existence of particular datasets stored in the mapset
            
            @note Will overwrite the value for a key that already exists
        
            @param context Context object containing projectDir, the path of the project whose 
            metadata store is to be written to
            @param key The key to be written to the GRASS section of the project metadata
            @param value The value to be written for key stored in the GRASS section of the project metadata
            
            @exception IOError(errno.EACCES) if the metadata store for the project is not writable
        """
        GenericMetadata.writeEntryToSection(context, GenericMetadata.GRASS_SECTION, key, value)
    
    
    @staticmethod 
    def writeClimatePointEntry(context, key, value):
        """ Write a point climate entry to the metadata store for a given project.
            
            @note Will overwrite the value for a key that already exists
        
            @param context Context object containing projectDir, the path of the project whose 
            metadata store is to be written to
            @param key The key to be written to the point climate section of the project metadata
            @param value The value to be written for key stored in the point climate section of the project metadata
            
            @exception IOError(errno.EACCES) if the metadata store for the project is not writable
        """
        GenericMetadata.writeEntryToSection(context, GenericMetadata.CLIMATE_POINT_SECTION, key, value)
        
        
    @staticmethod 
    def writeClimatePointEntries(context, keys, values):
        """ Write a point climate entries to the metadata store for a given project.
            
            @note Will overwrite the value for keys that already exist
        
            @param context Context object containing projectDir, the path of the project whose 
            metadata store is to be written to
            @param keys List of keys to be written to the point climate section of the project metadata
            @param values List of values to be written for keys stored in the point climate section of the project metadata
            
            @exception IOError(errno.EACCES) if the metadata store for the project is not writable
            @exception Exception if len(keys) != len(values)
        """
        GenericMetadata._writeEntriesToSection(context.projectDir, GenericMetadata.CLIMATE_POINT_SECTION, keys, values)
    
    
    @staticmethod 
    def writeClimateGridEntry(context, key, value):
        """ Write a grid climate entry to the metadata store for a given project.
            
            @note Will overwrite the value for a key that already exists
        
            @param context Context object containing projectDir, the path of the project whose 
            metadata store is to be read from
            @param key The key to be written to the grid climate  section of the project metadata
            @param value The value to be written for key stored in the grid climate  section of the project metadata
            
            @exception IOError(errno.EACCES) if the metadata store for the project is not writable
        """
        GenericMetadata.writeEntryToSection(context, GenericMetadata.CLIMATE_GRID_SECTION, key, value)
        
    
    @staticmethod
    def deleteManifestEntry(context, key):
        """ Delete an entry from the manifest section of the metadata store for a given project.
        
            @param context Context object containing projectDir, the path of the project whose 
            metadata store is to be deleted from
            @param key The key to be deleted from the given section of the project metadata
            
            @exception IOError(errno.EACCES) if the metadata store for the project is not writable
        """
        GenericMetadata.deleteEntryFromSection(context, GenericMetadata.MANIFEST_SECTION, key)
     
        
    @staticmethod 
    def deleteStudyAreaEntry(context, key):
        """ Delete an entry from the study area section of the metadata store for a given project.
        
            @param context Context object containing projectDir, the path of the project whose 
            metadata store is to be deleted from
            @param key The key to be deleted from the given section of the project metadata
            
            @exception IOError(errno.EACCES) if the metadata store for the project is not writable
        """
        GenericMetadata.deleteEntryFromSection(context, GenericMetadata.STUDY_AREA_SECTION, key)
    
    
    @staticmethod 
    def deleteGRASSEntry(context, key):
        """ Delete an entry from the GRASS section of the metadata store for a given project.
        
            @param context Context object containing projectDir, the path of the project whose 
            metadata store is to be deleted from
            @param key The key to be deleted from the given section of the project metadata
            
            @exception IOError(errno.EACCES) if the metadata store for the project is not writable
        """
        GenericMetadata.deleteEntryFromSection(context, GenericMetadata.GRASS_SECTION, key)
    
    
    @staticmethod 
    def deleteClimatePointEntry(context, key):
        """ Delete an entry from the point climate section of the metadata store for a given project.
        
            @param context Context object containing projectDir, the path of the project whose 
            metadata store is to be deleted from
            @param key The key to be deleted from the given section of the project metadata
            
            @exception IOError(errno.EACCES) if the metadata store for the project is not writable
        """
        GenericMetadata.deleteEntryFromSection(context, GenericMetadata.CLIMATE_POINT_SECTION, key)
        
    
    @staticmethod 
    def deleteClimateGridEntry(context, key):
        """ Delete an entry from the grid climate section of the metadata store for a given project.
        
            @param context Context object containing projectDir, the path of the project whose 
            metadata store is to be deleted from
            @param key The key to be deleted from the given section of the project metadata
            
            @exception IOError(errno.EACCES) if the metadata store for the project is not writable
        """
        GenericMetadata.deleteEntryFromSection(context, GenericMetadata.CLIMATE_GRID_SECTION, key)
    
        
    @staticmethod 
    def writeClimateGridEntries(context, keys, values):
        """ Write grid climate entries to the metadata store for a given project.
            
            @note Will overwrite the value for keys that already exist
        
            @param context Context object containing projectDir, the path of the project whose 
            metadata store is to be read from
            @param key List of keys to be written to the grid climate  section of the project metadata
            @param value List of values to be written for keys stored in the grid climate  section of the project metadata
            
            @exception IOError(errno.EACCES) if the metadata store for the project is not writable
            @exception Exception if len(keys) != len(values)
        """
        GenericMetadata._writeEntriesToSection(context.projectDir, GenericMetadata.CLIMATE_GRID_SECTION, keys, values)
    
    
    @staticmethod 
    def writeProvenanceEntry(context, key, value):
        """ Write a provenance entry to the metadata store for a given project.
            
            @note Will overwrite a the value for a key that already exists
        
            @param context Context object containing projectDir, the path of the project whose 
            metadata store is to be written to
            @param key The key to be written to the provenance section of the project metadata
            @param value The value to be written for key stored in the provenance section of the project metadata
            
            @exception IOError(errno.EACCES) if the metadata store for the project is not writable
        """
        GenericMetadata.writeEntryToSection(context, GenericMetadata.PROVENANCE_SECTION, key, value)
    
    
    @staticmethod 
    def writeModelRunEntries(context, keys, values):
        """ Write model run entries to the metadata store for a given project.
            
            @note Will overwrite the values of keys that already exist
        
            @param context Context object containing projectDir, the path of the project whose 
            metadata store is to be written to
            @param keys The keys to be written to the model run section of the project metadata
            @param values The values to be written for keys stored in the model run section of the project metadata
            
            @exception IOError(errno.EACCES) if the metadata store for the project is not writable
            @exception Exception if len(keys) != len(values)
        """
        GenericMetadata._writeEntriesToSection(context.projectDir, GenericMetadata.MODEL_RUN_SECTION, keys, values)
    
    
    @staticmethod 
    def writeProvenanceEntries(context, keys, values):
        """ Write provenance entries to the metadata store for a given project.
            
            @note Will overwrite the values of keys that already exist
        
            @param context Context object containing projectDir, the path of the project whose 
            metadata store is to be written to
            @param keys The keys to be written to the provenance section of the project metadata
            @param values The values to be written for keys stored in the provenance section of the project metadata
            
            @exception IOError(errno.EACCES) if the metadata store for the project is not writable
            @exception Exception if len(keys) != len(values)
        """
        GenericMetadata._writeEntriesToSection(context.projectDir, GenericMetadata.PROVENANCE_SECTION, keys, values)
    
    
    @staticmethod
    def _readEntriesForSection(projectDir, section):
        """ Read all entries for the given section from the metadata store for a given project
        
            @param projectDir Absolute path of the project whose metadata are to be read
            @param section The section the key is to be written to
            
            @return A dictionary of key/value pairs from the given section of the project metadata
        """
        sectionDict = dict()
        metadataFilepath = os.path.join(projectDir, GenericMetadata.METADATA_FILENAME)
        if os.path.exists(metadataFilepath):
            if not os.access(metadataFilepath, os.R_OK):
                raise IOError(errno.EACCES, "Unable to read metadata store for project %s" % \
                              (projectDir,))
            # Read metadata store
            config = ConfigParser.RawConfigParser()
            config.read(metadataFilepath)
            if config.has_section(section):
                items = config.items(section)
                for item in items:
                    sectionDict[item[0]] = item[1]
        
        return sectionDict
    
    
    @staticmethod
    def readManifestEntries(context):
        """ Read all manifest entries from the metadata store for a given project
        
            @param context Context object containing projectDir, the path of the project whose 
            metadata store is to be read from
            @return A dictionary of key/value pairs from the manifest section of the project metadata
        """
        return GenericMetadata._readEntriesForSection(context.projectDir, GenericMetadata.MANIFEST_SECTION)
    
    
    @staticmethod
    def readStudyAreaEntries(context):
        """ Read all study area entries from the metadata store for a given project
        
            @param context Context object containing projectDir, the path of the project whose 
            metadata store is to be read from
            
            @return A dictionary of key/value pairs from the study area section of the project metadata
        """
        return GenericMetadata._readEntriesForSection(context.projectDir, GenericMetadata.STUDY_AREA_SECTION)
    
    
    @staticmethod
    def readGRASSEntries(context):
        """ Read all GRASS entries from the metadata store for a given project
        
            @param context Context object containing projectDir, the path of the project whose 
            metadata store is to be read from
            
            @return A dictionary of key/value pairs from the GRASS section of the project metadata
        """
        return GenericMetadata._readEntriesForSection(context.projectDir, GenericMetadata.GRASS_SECTION)
    
    
    @staticmethod
    def readModelRunEntries(context):
        """ Read all point model run entries from the metadata store for a given project
        
            @param context Context object containing projectDir, the path of the project whose 
            metadata store is to be read from
            @return A dictionary of key/value pairs from the model run section of the project metadata
        """
        return GenericMetadata._readEntriesForSection(context.projectDir, GenericMetadata.MODEL_RUN_SECTION)
    
    
    @staticmethod
    def readModelRuns(context):
        """ Read all model runs from metadata and store in ModelRun 
            instances.
            
            @param context Context object containing projectDir, the path of the project whose 
            metadata store is to be read from
            @return A list of ModelRun objects
        """
        modelRunObjects = []
        modelRuns = GenericMetadata.readModelRunEntries(context)
        try:
            runs = modelRuns['runs'].split(GenericMetadata.VALUE_DELIM)
            for run in runs:
                modelRunObjects.append(ModelRun.readFromMetadata(context, run))
        except KeyError:
            pass
        return modelRunObjects

    
    @staticmethod
    def readClimatePointEntries(context):
        """ Read all point climate entries from the metadata store for a given project
        
            @param context Context object containing projectDir, the path of the project whose 
            metadata store is to be read from
            @return A dictionary of key/value pairs from the point climate section of the project metadata
        """
        return GenericMetadata._readEntriesForSection(context.projectDir, GenericMetadata.CLIMATE_POINT_SECTION)
    
    
    @staticmethod
    def readClimatePointStations(context):
        """ Read all climate point stations from metadata and store in ClimatePointStation 
            instances.
            
            @param context Context object containing projectDir, the path of the project whose 
            metadata store is to be read from
            @return A list of ClimatePointStation objects
        """
        stationObjects = []
        climatePoints = GenericMetadata.readClimatePointEntries(context)
        try:
            stations = climatePoints['stations'].split(GenericMetadata.VALUE_DELIM)
            for station in stations:
                stationObjects.append(ClimatePointStation.readFromMetadata(context, station))
        except KeyError:
            pass
        return stationObjects
    
    
    @staticmethod
    def readClimateGridEntries(context):
        """ Read all grid climate entries from the metadata store for a given project
        
            @param context Context object containing projectDir, the path of the project whose 
            metadata store is to be read from
            
            @return A dictionary of key/value pairs from the grid climate section of the project metadata
        """
        return GenericMetadata._readEntriesForSection(context.projectDir, GenericMetadata.CLIMATE_GRID_SECTION)


    @staticmethod
    def readProvenanceEntries(context):
        """ Read all provenance entries from the metadata store for a given project
        
            @param context Context object containing projectDir, the path of the project whose 
            metadata store is to be read from
            
            @return A dictionary of key/value pairs from the provenance section of the project metadata
        """
        return GenericMetadata._readEntriesForSection(context.projectDir, GenericMetadata.PROVENANCE_SECTION)


    @staticmethod
    def readAssetProvenanceObjects(context):
        """ Read all asset provenance objects from metadata and store in AssetProvenance
            instances.
            
            @param context Context object containing projectDir, the path of the project whose 
            metadata store is to be read from
            
            @return A list of AssetProvenance objects
        """
        assetProvenanceObjects = []
        provenance = GenericMetadata.readProvenanceEntries(context)
        try:
            assets = provenance['entities'].split(GenericMetadata.VALUE_DELIM)
            for asset in assets:
                assetProvenanceObjects.append(AssetProvenance.readFromMetadata(context, asset))
        except KeyError:
            pass
        return assetProvenanceObjects
    
    
    @staticmethod
    def getProcessingHistoryList(context):
        """ Get processing history stored in the project metadata
        
            @param context Context object containing projectDir, the path of the project whose 
            metadata store is to be written to
            @return List containing strings representing history items
        """
        projectDir = context.projectDir
        steps = []
        history = GenericMetadata._readEntriesForSection(projectDir, GenericMetadata.HISTORY_SECTION)
        try:
            idx = int(history['numsteps']) + 1
            for i in xrange(1, idx):
                key = GenericMetadata.HISTORY_PROTO + str(i)
                steps.append(history[key])
        except KeyError:
            pass
        
        return steps
    
    @staticmethod
    def appendProcessingHistoryItem(context, item):
        """ Write an item to the processing history stored in the project metadata
        
            @param context Context object containing projectDir, the path of the project whose 
            metadata store is to be written to
            @param item String representing item to be written to processing history
        """
        projectDir = context.projectDir
        history = GenericMetadata._readEntriesForSection(projectDir, GenericMetadata.HISTORY_SECTION)
        try:
            idx = int(history['numsteps'])
        except KeyError:
            idx = 0
        idx += 1
        
        idxStr = str(idx)
        key = GenericMetadata.HISTORY_PROTO + idxStr
        GenericMetadata._writeEntriesToSection(projectDir, GenericMetadata.HISTORY_SECTION, [key, 'numsteps'], [item, idxStr])
        
        


"""@package ecohydrolib.tests.test_metadata
    
    @brief Test methods for ecohydrolib.metadata
    
    This software is provided free of charge under the New BSD License. Please see
    the following license information:
    
    Copyright (c) 2013, University of North Carolina at Chapel Hill
    All rights reserved.
    
    Redistribution and use in source and binary forms, with or without
    modification, are permitted provided that the following conditions are met:
        * Redistributions of source code must retain the above copyright
          notice, this list of conditions and the following disclaimer.
        * Redistributions in binary form must reproduce the above copyright
          notice, this list of conditions and the following disclaimer in the
          documentation and/or other materials provided with the distribution.
        * Neither the name of the University of North Carolina at Chapel Hill nor the
          names of its contributors may be used to endorse or promote products
          derived from this software without specific prior written permission.
    
    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
    ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
    WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
    DISCLAIMED. IN NO EVENT SHALL THE UNIVERSITY OF NORTH CAROLINA AT CHAPEL HILL
    BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
    CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
    GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
    HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
    LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
    OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


    @author Brian Miles <brian_miles@unc.edu>
    
    Usage: 
    @code
    python -m unittest test_metadata
    @endcode
    
""" 
from unittest import TestCase
import os
from datetime import datetime
import tempfile, shutil

from ecohydrolib.context import Context
from ecohydrolib.metadata import GenericMetadata
from ecohydrolib.metadata import ClimatePointStation
from ecohydrolib.metadata import AssetProvenance
from ecohydrolib.metadata import MetadataVersionError

class TestMetadata(TestCase):
    
    def setUp(self):
        self.projectDir = tempfile.mkdtemp()
        self.testMetadataPath = os.path.join(self.projectDir, GenericMetadata.METADATA_FILENAME)
        if os.path.exists(self.testMetadataPath):
            os.unlink(self.testMetadataPath)
        lockFilePath = os.path.join(self.projectDir, GenericMetadata.METADATA_LOCKFILE)
        if os.path.exists(lockFilePath):
            os.unlink(lockFilePath)
        self.context = Context(projectDir=self.projectDir)
        
    def tearDown(self):
        if os.path.exists(self.testMetadataPath):
            os.unlink(self.testMetadataPath)
        lockFilePath = os.path.join("/tmp", GenericMetadata.METADATA_LOCKFILE)
        if os.path.exists(lockFilePath):
            os.unlink(lockFilePath)
        shutil.rmtree(self.projectDir)
    
    
    def test_empty_read(self):
        manifest = GenericMetadata.readManifestEntries(self.context)
        self.assertTrue(len(manifest) == 0)
        
        studyArea = GenericMetadata.readStudyAreaEntries(self.context)
        self.assertTrue(len(studyArea) == 0)
    
        climatePoint = GenericMetadata.readClimatePointEntries(self.context)
        self.assertTrue(len(climatePoint) == 0)
        
        climateGrid = GenericMetadata.readClimateGridEntries(self.context)
        self.assertTrue(len(climateGrid) == 0)
        
        
    def test_write_and_read(self):
        GenericMetadata.writeManifestEntry(self.context, "key1", "value_one")
        GenericMetadata.writeManifestEntry(self.context, "key2", "value_two")
        manifest = GenericMetadata.readManifestEntries(self.context)
        self.assertTrue(manifest["key1"] == "value_one")
        
        GenericMetadata.writeStudyAreaEntry(self.context, "key1", "value_one")
        GenericMetadata.writeStudyAreaEntry(self.context, "key2", "value_two")
        studyArea = GenericMetadata.readStudyAreaEntries(self.context)
        self.assertTrue(studyArea["key1"] == "value_one")
        
        GenericMetadata.writeClimatePointEntry(self.context, "key1", "value_one")
        GenericMetadata.writeClimatePointEntry(self.context, "key2", "value_two")
        climatePoint = GenericMetadata.readClimatePointEntries(self.context)
        self.assertTrue(climatePoint["key1"] == "value_one")
        
        GenericMetadata.writeClimateGridEntry(self.context, "key1", "value_one")
        GenericMetadata.writeClimateGridEntry(self.context, "key2", "value_two")
        climateGrid = GenericMetadata.readClimateGridEntries(self.context)
        self.assertTrue(climateGrid["key1"] == "value_one")
        
        
    def test_delete(self):
        GenericMetadata.writeManifestEntry(self.context, "key1", "value_one")
        manifest = GenericMetadata.readManifestEntries(self.context)
        self.assertTrue(manifest["key1"] == "value_one")
        GenericMetadata.deleteManifestEntry(self.context, "key1")
        manifest = GenericMetadata.readManifestEntries(self.context)
        self.assertTrue(not 'key1' in manifest)
        
        GenericMetadata.writeStudyAreaEntry(self.context, "key1", "value_one")
        studyArea = GenericMetadata.readStudyAreaEntries(self.context)
        self.assertTrue(studyArea["key1"] == "value_one")
        GenericMetadata.deleteStudyAreaEntry(self.context, 'key1')
        studyArea = GenericMetadata.readStudyAreaEntries(self.context)
        self.assertTrue(not 'key1' in studyArea)
        
        GenericMetadata.writeClimatePointEntry(self.context, "key1", "value_one")
        climatePoint = GenericMetadata.readClimatePointEntries(self.context)
        self.assertTrue(climatePoint["key1"] == "value_one")
        GenericMetadata.deleteClimatePointEntry(self.context, 'key1')
        climatePoint = GenericMetadata.readClimatePointEntries(self.context)
        self.assertTrue(not 'key1' in climatePoint)
        
        GenericMetadata.writeClimateGridEntry(self.context, "key1", "value_one")
        climateGrid = GenericMetadata.readClimateGridEntries(self.context)
        self.assertTrue(climateGrid["key1"] == "value_one")
        GenericMetadata.deleteClimateGridEntry(self.context, 'key1')
        climateGrid = GenericMetadata.readClimateGridEntries(self.context)
        self.assertTrue(not 'key1' in climateGrid)
        # Delete and empty entry
        GenericMetadata.deleteClimateGridEntry(self.context, "not_in_store")
        
        
    def test_write_climate_point1(self):
        """ Test case where there is a single data file the station """
        station = ClimatePointStation()
        station.type = "GHCN"
        station.id = "US1MDBL0027"
        station.longitude = -76.716
        station.latitude = 39.317
        station.elevation = 128.0
        station.name = "My station name"
        station.data = "clim.txt"
        station.startDate = datetime.strptime("201007", "%Y%m")
        station.endDate = datetime.strptime("201110", "%Y%m")
        station.variables = [ClimatePointStation.VAR_PRECIP, \
                             ClimatePointStation.VAR_SNOW]
        station.writeToMetadata(self.context)
        
        climatePointStation = GenericMetadata.readClimatePointStations(self.context)[0]  
        self.assertTrue(station.type.lower() == climatePointStation.type)
        self.assertTrue(station.id.lower() == climatePointStation.id)
        self.assertTrue(station.longitude == climatePointStation.longitude)
        self.assertTrue(station.latitude == climatePointStation.latitude)
        self.assertTrue(station.elevation == climatePointStation.elevation)
        self.assertTrue(station.name == climatePointStation.name)
        self.assertTrue(station.data == climatePointStation.data)
        self.assertTrue(station.startDate == climatePointStation.startDate)
        self.assertTrue(station.endDate == climatePointStation.endDate)
        self.assertTrue(station.variables == climatePointStation.variables)
        
    def test_write_climate_point1_overwrite(self):
        """ Test case where there is a single data file the station, the entry is overwritten """
        station = ClimatePointStation()
        station.type = "GHCN"
        station.id = "US1MDBL0027"
        station.longitude = -76.716
        station.latitude = 39.317
        station.elevation = 128.0
        station.name = "My station name"
        station.data = "clim.txt"
        station.startDate = datetime.strptime("201007", "%Y%m")
        station.endDate = datetime.strptime("201110", "%Y%m")
        station.variables = [ClimatePointStation.VAR_PRECIP, \
                             ClimatePointStation.VAR_SNOW]
        station.writeToMetadata(self.context)
        
        climatePointStation = GenericMetadata.readClimatePointStations(self.context)[0]  
        self.assertTrue(station.type.lower() == climatePointStation.type)
        self.assertTrue(station.id.lower() == climatePointStation.id)
        self.assertTrue(station.longitude == climatePointStation.longitude)
        self.assertTrue(station.latitude == climatePointStation.latitude)
        self.assertTrue(station.elevation == climatePointStation.elevation)
        self.assertTrue(station.name == climatePointStation.name)
        self.assertTrue(station.data == climatePointStation.data)
        self.assertTrue(station.startDate == climatePointStation.startDate)
        self.assertTrue(station.endDate == climatePointStation.endDate)
        self.assertTrue(station.variables == climatePointStation.variables)
        
        station.longitude = -76.716
        station.latitude = 39.317
        station.elevation = 128.0
        station.name = "My (longer) station name"
        station.data = "clim.dat"
        station.startDate = datetime.strptime("201006", "%Y%m")
        station.endDate = datetime.strptime("201310", "%Y%m")
        station.variables = [ClimatePointStation.VAR_PRECIP, \
                             ClimatePointStation.VAR_SNOW]
        station.writeToMetadata(self.context)
        
        climatePointStation = GenericMetadata.readClimatePointStations(self.context)[0]  
        self.assertTrue(station.type.lower() == climatePointStation.type)
        self.assertTrue(station.id.lower() == climatePointStation.id)
        self.assertTrue(station.longitude == climatePointStation.longitude)
        self.assertTrue(station.latitude == climatePointStation.latitude)
        self.assertTrue(station.elevation == climatePointStation.elevation)
        self.assertTrue(station.name == climatePointStation.name)
        self.assertTrue(station.data == climatePointStation.data)
        self.assertTrue(station.startDate == climatePointStation.startDate)
        self.assertTrue(station.endDate == climatePointStation.endDate)
        self.assertTrue(station.variables == climatePointStation.variables)
        
        
    def test_write_climate_point2(self):
        """ Test case where there are separate data files for each variable and there are two climate stations """
        station = ClimatePointStation()
        station.type = "GHCN"
        station.id = "US1MDBL0027"
        station.longitude = -76.716
        station.latitude = 39.317
        station.elevation = 128.0
        station.name = "My station name"
        station.startDate = datetime.strptime("201007", "%Y%m")
        station.endDate = datetime.strptime("201110", "%Y%m")
        station.variables = [ClimatePointStation.VAR_PRECIP, \
                             ClimatePointStation.VAR_SNOW]
        station.variablesData[ClimatePointStation.VAR_PRECIP] = ClimatePointStation.VAR_PRECIP + '.txt'
        station.variablesData[ClimatePointStation.VAR_SNOW] = ClimatePointStation.VAR_SNOW + '.txt'
        station.writeToMetadata(self.context)
        
        station2 = ClimatePointStation()
        station2.type = "GHCN"
        station2.id = "US1MDBL4242"
        station2.longitude = -42.716
        station2.latitude = 42.317
        station2.elevation = 42.0
        station2.name = "My 42 station"
        station2.startDate = datetime.strptime("199907", "%Y%m")
        station2.endDate = datetime.strptime("200110", "%Y%m")
        station2.variables = [ClimatePointStation.VAR_PRECIP, \
                             ClimatePointStation.VAR_SNOW]
        station2.variablesData[ClimatePointStation.VAR_PRECIP] = ClimatePointStation.VAR_PRECIP + '.txt'
        station2.variablesData[ClimatePointStation.VAR_SNOW] = ClimatePointStation.VAR_SNOW + '.txt'
        station2.writeToMetadata(self.context)
        
        climatePointStation = GenericMetadata.readClimatePointStations(self.context)[0]        
        self.assertTrue(station.type.lower() == climatePointStation.type)
        self.assertTrue(station.id.lower() == climatePointStation.id)
        self.assertTrue(station.longitude == climatePointStation.longitude)
        self.assertTrue(station.latitude == climatePointStation.latitude)
        self.assertTrue(station.elevation == climatePointStation.elevation)
        self.assertTrue(station.name == climatePointStation.name)
        self.assertTrue(station.startDate == climatePointStation.startDate)
        self.assertTrue(station.endDate == climatePointStation.endDate)
        self.assertTrue(station.variables == climatePointStation.variables)
        self.assertTrue(station.variablesData[ClimatePointStation.VAR_PRECIP] == climatePointStation.variablesData[ClimatePointStation.VAR_PRECIP])
        self.assertTrue(station.variablesData[ClimatePointStation.VAR_SNOW] == climatePointStation.variablesData[ClimatePointStation.VAR_SNOW])
        
    def test_provenance(self):
        """ Test case writing provenance metadata """
        asset = AssetProvenance()
        asset.section = GenericMetadata.MANIFEST_SECTION
        asset.name = "dem"
        asset.dcIdentifier = "dem.tif"
        asset.dcSource = "http://www.demexplorer.com/..."
        asset.dcTitle = "Study area DEM"
        asset.dcDate = datetime.strptime("201303", "%Y%m")
        asset.dcPublisher = "USGS"
        asset.dcDescription = "RegisterDEM.py ..."
        asset.writeToMetadata(self.context)
        
        assetProvenance = GenericMetadata.readAssetProvenanceObjects(self.context)[0]
        self.assertTrue(asset.section == assetProvenance.section)
        self.assertTrue(asset.name == assetProvenance.name)
        self.assertTrue(asset.dcIdentifier == assetProvenance.dcIdentifier)
        self.assertTrue(asset.dcSource == assetProvenance.dcSource)
        self.assertTrue(asset.dcTitle == assetProvenance.dcTitle)
        self.assertTrue(asset.dcDate == assetProvenance.dcDate)
        self.assertTrue(asset.dcPublisher == assetProvenance.dcPublisher)
        self.assertTrue(asset.dcDescription == assetProvenance.dcDescription)
        
    def test_provenance_overwrite(self):
        """ Test case writing provenance metadata, with overwrite """
        asset = AssetProvenance()
        asset.section = GenericMetadata.MANIFEST_SECTION
        asset.name = "dem"
        asset.dcIdentifier = "dem.tif"
        asset.dcSource = "http://www.demexplorer.com/..."
        asset.dcTitle = "Study area DEM"
        asset.dcDate = datetime.strptime("201303", "%Y%m")
        asset.dcPublisher = "USGS"
        asset.dcDescription = "RegisterDEM.py ..."
        asset.writeToMetadata(self.context)
        
        assetProvenance = GenericMetadata.readAssetProvenanceObjects(self.context)[0]
        self.assertTrue(asset.section == assetProvenance.section)
        self.assertTrue(asset.name == assetProvenance.name)
        self.assertTrue(asset.dcIdentifier == assetProvenance.dcIdentifier)
        self.assertTrue(asset.dcSource == assetProvenance.dcSource)
        self.assertTrue(asset.dcTitle == assetProvenance.dcTitle)
        self.assertTrue(asset.dcDate == assetProvenance.dcDate)
        self.assertTrue(asset.dcPublisher == assetProvenance.dcPublisher)
        self.assertTrue(asset.dcDescription == assetProvenance.dcDescription)
        
        asset.dcIdentifier = 'foo.img'
        asset.dcSource = "http://a.different.url/..."
        asset.dcTitle = "A different study area DEM"
        asset.dcDate = datetime.strptime("201304", "%Y%m")
        asset.dcPublisher = "NASA"
        asset.dcDescription = "GetDEMExplorerDEM.py ..."
        asset.writeToMetadata(self.context)
        
        assetProvenance = GenericMetadata.readAssetProvenanceObjects(self.context)[0]
        self.assertTrue(asset.section == assetProvenance.section)
        self.assertTrue(asset.name == assetProvenance.name)
        self.assertTrue(asset.dcIdentifier == assetProvenance.dcIdentifier)
        self.assertTrue(asset.dcSource == assetProvenance.dcSource)
        self.assertTrue(asset.dcTitle == assetProvenance.dcTitle)
        self.assertTrue(asset.dcDate == assetProvenance.dcDate)
        self.assertTrue(asset.dcPublisher == assetProvenance.dcPublisher)
        self.assertTrue(asset.dcDescription == assetProvenance.dcDescription)
        
        
    def test_processing_history(self):
        """ Test processing history metadata """
        projectDir = "/tmp"
        
        step1 = "mkdir foo; cd foo"
        step2 = "touch README.txt"
        step3 = "git init"
        
        GenericMetadata.appendProcessingHistoryItem(self.context, step1)
        GenericMetadata.appendProcessingHistoryItem(self.context, step2)
        GenericMetadata.appendProcessingHistoryItem(self.context, step3)
        
        history = GenericMetadata.getProcessingHistoryList(self.context)
        self.assertTrue(len(history) == 3, "Expected history length to be 3, but it is %d" % (len(history),) )
        self.assertTrue(history[0] == step1)
        self.assertTrue(history[1] == step2)
        self.assertTrue(history[2] == step3)
        
    def test_version_conflict(self):
        """ Induce a version conflict """
        
        step1 = "mkdir foo; cd foo"
        step2 = "touch README.txt"
        
        GenericMetadata.appendProcessingHistoryItem(self.context, step1)
        # For testing purposes only, users should not modify 
        #   GenericMetadata._ecohydrolibVersion
        _prevVersion = GenericMetadata._ecohydrolibVersion
        GenericMetadata._ecohydrolibVersion = '11'
        caughtMetadataVersionError = False
        try:
            GenericMetadata.appendProcessingHistoryItem(self.context, step2)
        except MetadataVersionError:
            caughtMetadataVersionError = True
        self.assertTrue(caughtMetadataVersionError, "Expected metadata version mismatch, but none found.")
        GenericMetadata._ecohydrolibVersion = _prevVersion
        
    def test_check_metadata_version(self):
        """ Test check metadata version """       
        step1 = "mkdir foo; cd foo"
        
        GenericMetadata.appendProcessingHistoryItem(self.context, step1)
        GenericMetadata.checkMetadataVersion(self.context.projectDir)
        # For testing purposes only, users should not modify 
        #   GenericMetadata._ecohydrolibVersion
        _prevVersion = GenericMetadata._ecohydrolibVersion
        GenericMetadata._ecohydrolibVersion = '11'
        caughtMetadataVersionError = False
        try:
            GenericMetadata.checkMetadataVersion(self.context.projectDir)
        except MetadataVersionError:
            caughtMetadataVersionError = True
        self.assertTrue(caughtMetadataVersionError, "Expected metadata version mismatch, but none found.")
        GenericMetadata._ecohydrolibVersion = _prevVersion
        
        


"""@package ecohydrolib.ssurgo.rasterize
    
@brief Rasterize SSURGO features

This software is provided free of charge under the New BSD License. Please see
the following license information:

Copyright (c) 2013, University of North Carolina at Chapel Hill
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of the University of North Carolina at Chapel Hill nor the
      names of its contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE UNIVERSITY OF NORTH CAROLINA AT CHAPEL HILL
BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


@author Brian Miles <brian_miles@unc.edu>
"""
import os.path, errno
import osr

from ecohydrolib.spatialdata.utils import deleteGeoTiff
from ecohydrolib.spatialdata.utils import getSpatialReferenceForRaster
from ecohydrolib.spatialdata.utils import getMeterConversionFactorForLinearUnitOfShapefile
from ecohydrolib.spatialdata.utils import getMeterConversionFactorForLinearUnitOfGMLfile

import attributequery

RASTER_ATTRIBUTES = attributequery.ATTRIBUTE_LIST_NUMERIC
# Depth to bed rock from MapunitPolyExtended
RASTER_ATTRIBUTES.append('brockdepmin')


def deleteSoilRasters(context, manifest):
    """ Delete soil raster maps stored in a project
    
        @param context Context object containing projectDir, the path of the project whose 
        metadata store is to be read from
        @param manifest Dict containing manifest entries.  Files associted with entries
        whose key begins with 'soil_raster_' will be deleted
    """
    for entry in manifest.keys():
        if entry.find('soil_raster_') == 0:
            filePath = os.path.join( context.projectDir, manifest[entry] )
            deleteGeoTiff(filePath)


def rasterizeSSURGOFeatures(config, outputDir, featureFilename, featureLayername, featureAttrList, \
                            getResolutionFromRasterFileNamed=None, rasterResolutionX=None, rasterResolutionY=None):
    """ Create raster maps, in GeoTIFF format, for SSURGO attributes associated with SSURGO MapunitPoly/MapunitPolyExtended features
        
        @note Will silently exit if rasters already exist.
        @note If getResolutionFromRasterFileNamed as well as rasterResolutionX and rasterResolutionY are specified,
        output raster resolution will be determined from the file named by getResolutionFromRasterFileNamed. 
        
        @param config onfigParser containing the section 'GDAL/OGR' and option 'PATH_OF_GDAL_RASTERIZE'
        @param outputDir String representing the absolute/relative path of the directory into which output rasters should be written
        @param featureFilename String representing the absolute/relative path of the input feature in ESRI shapefile or OGC GML format
        @param featureLayername String representing the name of feature layer whose features are to be rasterize
        @param featureAttrList List containing the SSURGO attributes for which raster maps are to be created
        @param getResolutionFromRasterFileNamed String representing the absolute path of an existing raster file from which the
            output raster resolution should be determined
        @param rasterResolutionX Float representing the X resolution of the output rasters
        @param rasterResolutionY Float representing the Y resolution of the output rasters
        
        @return Dictionary containing the keys for each soil attribute and values of the names of the raster files generated for that attribute
        
        @exception ConfigParser.NoSectionError
        @exception ConfigParser.NoOptionError
        @exception Exception if featureFilename is neither a SHP nor a GML file
        @exception IOError(errno.ENOTDIR) if outputDir is not a directory
        @exception IOError(errno.EACCESS) if outputDir is not writable
        @exception IOError(errno.EACCESS) if featureFilename is not readable
        @exception IOError(errno.ENOENT) if featureFilename is not a file
        @exception IOError(errno.EACCES) if getResolutionFromRasterFileNamed is not readable
        @exception ValueError if rasterResolutionX or rasterResolutionY are not floating point numbers greater than 0
        @exception Exception if a gdal_rasterize command fails
    """
    gdalCmdPath = config.get('GDAL/OGR', 'PATH_OF_GDAL_RASTERIZE')
    if not os.access(gdalCmdPath, os.X_OK):
        raise IOError(errno.EACCES, "The gdal_rasterize binary at %s is not executable" %
                      gdalCmdPath)
    gdalCmdPath = os.path.abspath(gdalCmdPath)
    
    if not os.path.isdir(outputDir):
        raise IOError(errno.ENOTDIR, "Output directory %s is not a directory" % (outputDir,))
    if not os.access(outputDir, os.W_OK):
        raise IOError(errno.EACCES, "Not allowed to write to output directory %s" % (outputDir,))
    outputDir = os.path.abspath(outputDir)
    
    featureFilepath = os.path.join(outputDir, featureFilename)
    if not os.access(featureFilepath, os.R_OK):
        raise IOError(errno.EACCES, "Not allowed to read feature file %s" % (featureFilepath,))
    if not os.path.isfile(featureFilepath):
        raise IOError(errno.ENOENT, "Feature file %s does not exist" % (featureFilepath,))
    
    if getResolutionFromRasterFileNamed is None:
        rasterResolutionX = float(rasterResolutionX)
        if rasterResolutionX <= 0.0:
            raise ValueError("rasterResolutionX must be > 0.0")
        rasterResolutionY = float(rasterResolutionY)
        if rasterResolutionY <= 0.0:
            raise ValueError("rasterResolutionY must be > 0.0")
    else:
        if not os.access(getResolutionFromRasterFileNamed, os.R_OK):
            raise IOError(errno.EACCES, "Not allowed to read existing raster file %s from which to get resolution" % \
                          (getResolutionFromRasterFileNamed))

    featureType = os.path.splitext(featureFilepath)[1].lstrip('.').upper()

    # Get conversion factor for converting soil feature linear unit into meters
    if featureType == "GML":
        soilMeterConvFactor = getMeterConversionFactorForLinearUnitOfGMLfile(featureFilepath)
    elif featureType == "SHP":
        soilMeterConvFactor = getMeterConversionFactorForLinearUnitOfShapefile(featureFilepath)
    else:
        raise Exception("Argument featureFilename must be either a GML or SHP file")
        
    # Determine soil property raster map resolution, in the same linear units as soil feature layer
    if getResolutionFromRasterFileNamed is not None:
        # Use DEM raster resolution as soil property raster resolution
        #sys.stderr.write("Using input raster resolution for soil property rasters ...\n")
        demPixelSizeAndUnits = getSpatialReferenceForRaster(getResolutionFromRasterFileNamed)
        # Default, DEM raster linear unit is the same as soil feature linear units
        rasterResolutionX = demPixelSizeAndUnits[0]
        rasterResolutionY = demPixelSizeAndUnits[1]
        # Get DEM raster linear units
        demSrs = osr.SpatialReference()
        if demSrs.ImportFromWkt(demPixelSizeAndUnits[4]) == 0: # OGRERR_NONE = 0
            #sys.stderr.write("Spatial reference found for DEM, converting DEM linear units to soil feature linear units\n")
            demMeterConvFactor = demSrs.GetLinearUnits()
            if soilMeterConvFactor != demMeterConvFactor:
                # Soil feature linear units differ from DEM linear units, convert to meters
                rasterResolutionX = (demPixelSizeAndUnits[0] * demMeterConvFactor)
                rasterResolutionY = (demPixelSizeAndUnits[1] * demMeterConvFactor)
        else:
            pass
            #sys.stderr.write("No spatial reference found for DEM, assuming DEM linear units are the same as soil feature linear units\n")
    else:
        # Use user-specified soil raster resolution, but convert from meters into soil feature linear units
        pass
        #sys.stderr.write("Using user-supplied resolution, %d meters x %d meters, for soil property rasters ...\n" % (rasterResolutionX, rasterResolutionY))
        
    # Convert output raster resolution from meters into the linear units of the soil feature shapefile 
    rasterResolutionX = rasterResolutionX / soilMeterConvFactor
    rasterResolutionY = rasterResolutionY / soilMeterConvFactor
    
    # Build base filename for all output raster maps
    rasterFilenameProto = os.path.splitext(featureFilename)[0]
    
    filesCreated = dict()
    
    # Rasterize soil feature attributes using gdal_rasterize
    for attr in featureAttrList:
        rasterFilename = "%s_%s.tif" % (rasterFilenameProto, attr)
        filesCreated[attr] = rasterFilename
        rasterFilepath = os.path.join(outputDir, rasterFilename)
        if not os.path.exists(rasterFilepath):
            gdalCommand = "%s -q -of GTiff -co 'COMPRESS=LZW' -tr %d %d -a_nodata -9999 -a %s -l %s %s %s" % (gdalCmdPath, rasterResolutionX, rasterResolutionY, attr, featureLayername, featureFilepath, rasterFilepath)
            #print gdalCommand
            returnCode = os.system(gdalCommand)
            if returnCode != 0:
                raise Exception("GDAL command %s failed.  Check spatial reference system of input vector dataset (geographic coordinate systems may not work)." % (gdalCommand,))
    return filesCreated  
        


"""@package ecohydrolib.nhdplus2.networkanalysis
    
@brief Methods for querying the NHDPlus V2 data set. Requires that a NHDPlus V2 
database be initialized from data archive files using NHDPlusSetup.py

This software is provided free of charge under the New BSD License. Please see
the following license information:

Copyright (c) 2013, University of North Carolina at Chapel Hill
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of the University of North Carolina at Chapel Hill nor the
      names of its contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE UNIVERSITY OF NORTH CAROLINA AT CHAPEL HILL
BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


@author Brian Miles <brian_miles@unc.edu>
"""
import os
import errno
import time
import sqlite3
import re

import ogr
from shapely.geometry import Polygon
from shapely.wkb import loads, dumps
from shapely.ops import *

from ecohydrolib.spatialdata.utils import OGR_SHAPEFILE_DRIVER_NAME
from ecohydrolib.spatialdata.utils import OGR_DRIVERS
from ecohydrolib.spatialdata.utils import getBoundingBoxForShapefile
from ecohydrolib.spatialdata.utils import deleteShapefile

OGR_UPDATE_MODE = False
NORTH = 0
EAST = 90
UPSTREAM_SEARCH_THRESHOLD = 998


def getNHDReachcodeAndMeasureForGageSourceFea(config, source_fea):
    """ Get NHD Reachcode and measure along reach for a 
        streamflow gage identified by a source_fea (e.g. USGS Site Number)
    
        @param config A Python ConfigParser containing the following sections and options:
            'NHDPLUS2', 'PATH_OF_NHDPLUS2_DB' (absolute path to NHDPlus2 SQLite3 database)
        @param source_fea String representing source_fea of GageLoc gage
         
        @return A tuple(string, float) representing the reachcode and measure; None if no gage was found.
        
        @raise ConfigParser.NoSectionError
        @raise ConfigParser.NoOptionError
        @raise IOError(errno.EACCES) if NHDPlus2 DB is not readable
    """
    nhddbPath = config.get('NHDPLUS2', 'PATH_OF_NHDPLUS2_DB')
    if not os.access(nhddbPath, os.R_OK):
        raise IOError(errno.EACCES, "The database at %s is not readable" %
                      nhddbPath)
    nhddbPath = os.path.abspath(nhddbPath)
    
    conn = sqlite3.connect(nhddbPath)
    
    cursor = conn.cursor()
    cursor.execute("""SELECT ReachCode,Measure FROM Gage_Loc WHERE Source_Fea=?""", (source_fea,))
    result = cursor.fetchone()
    if None == result:
        return None
    
    reachcode = result[0]
    measure = result[1]

    return (reachcode, measure)


def getLocationForStreamGageByGageSourceFea(config, source_fea):
    """ Get lat/lon, in WGS84 (EPSG:4326), from gage point layer (Gage_Loc) for
        gage identified by a source_fea (e.g. USGS Site Number)
    
        @param config A Python ConfigParser containing the following sections and options:
            'NHDPLUS2', 'PATH_OF_NHDPLUS2_GAGELOC' (absolute path to NHD GageLoc SQLite3 spatial database)
        @param source_fea String representing source_fea of GageLoc gage
         
        @return A tuple with (x,y) coordinates in 'EPSG:4326' (WGS 84)
    """
    whereFilter = "Source_Fea='%s'" % (source_fea,)
    result = getLocationForStreamGage(config, whereFilter)
    if result:
        return (result[0], result[1])
    return None


def getLocationForStreamGageByReachcodeAndMeasure(config, reachcode, measure):
    """ Get lat/lon, in WGS84 (EPSG:4326), from gage point layer (Gage_Loc) for
        gage identified by reachcode and measure
    
        @param config A Python ConfigParser containing the following sections and options:
            'NHDPLUS2', 'PATH_OF_NHDPLUS2_GAGELOC' (absolute path to NHD GageLoc SQLite3 spatial database)
        @param reachcode String representing NHD streamflow gage 
        @param measure Float representing the measure along reach where Stream Gage is located 
            in percent from downstream end of the one or more NHDFlowline features that are 
            assigned to the ReachCode (see NHDPlusV21 GageLoc table)
         
        @return A tuple with (x,y) coordinates in 'EPSG:4326' (WGS 84); None if gage was not found
    """
    whereFilter = "REACHCODE='%s' and Measure=%f" % (reachcode, measure)
    result = getLocationForStreamGage(config, whereFilter)
    if result:
        return (result[0], result[1])
    return None


def getLocationForStreamGage(config, whereFilter):
    """ Get lat/lon, in WGS84 (EPSG:4326), from gage point layer (Gage_Loc) for
        gage identified by reachcode and measure
    
        @param config A Python ConfigParser containing the following sections and options:
            'NHDPLUS2', 'PATH_OF_NHDPLUS2_GAGELOC' (absolute path to NHD GageLoc SQLite3 spatial database)
        @param whereFilter String representing the whereFilter to use
         
        @return A tuple with (x,y) coordinates in 'EPSG:4326' (WGS 84); None if gage was not found
        
        @raise ConfigParser.NoSectionError
        @raise ConfigParser.NoOptionError
        @raise Exception if unable to open gage database
        @raise IOError(errno.ENOTDIR) if GageLoc is not readable
    """
    gageLocPath = config.get('NHDPLUS2', 'PATH_OF_NHDPLUS2_GAGELOC')
    if not os.access(gageLocPath, os.R_OK):
        raise IOError(errno.EACCES, "The database at %s is not readable" %
                      gageLocPath)
    gageLocPath = os.path.abspath(gageLocPath)
    
    poDS = ogr.Open(gageLocPath, OGR_UPDATE_MODE)
    if not poDS:
        raise Exception("Unable to open gage database %s" (gageLocPath,))
    assert(poDS.GetLayerCount() > 0)
    poLayer = poDS.GetLayer(0)
    assert(poLayer)
    assert(poLayer.SetAttributeFilter(whereFilter) == 0)
    poFeature = poLayer.GetNextFeature()
    if poFeature:
        poGeometry = poFeature.GetGeometryRef()
        
        # Make sure spatial reference is EPSG:4326
        pattern = re.compile("\ssrsName=\"(.+)\">")
        gml = poGeometry.ExportToGML()
        result = pattern.search(gml)
        assert(result.group(1) == 'EPSG:4326')
        
        # Get coordinates
        x = poGeometry.GetX()
        y = poGeometry.GetY()
        return (x,y)
    return None
    

def getComIdForStreamGage(conn, reachcode, measure):
    """ Uses NHDFlowline and/or NHDReachCode_ComID table(s) to lookup the ComID associated with a stream gage
        identified by reach code and measure.
    
        @param conn An sqlite3 connection to a database that has the NHDFlowline and NHDReachCode_Comid tables
        @param reachcode An string representing the Reachcode
        #param measure A float representing the measure along reach where Stream Gage is located 
            in percent from downstream end of the one or more NHDFlowline features that are 
            assigned to the ReachCode (see NHDPlusV21 GageLoc table)
        
        @return An integer representing the ComID associated with a Reachcode, or -1 if a reach with Reachcode 
        was not found.

    """
    comID = -1
    cursor = conn.cursor()
    # NHDPlusV21:
    ## The NHDFlowline comid for a stream flow gage location can be determined from the 
    ## PlusFlowlineVAA where Gage_Loc.Reachcode = PlusFlowlineVAA.Reachcode and 
    ## Gage_Loc.measure => PlusFlowlineVAA.FromMeas and Gage_Loc.measure <= PlusFlowlineVAA.ToMeas.
    cursor.execute("""SELECT p.ComID FROM PlusflowlineVAA as p
JOIN Gage_Loc as g ON p.ReachCode=g.ReachCode
WHERE (g.Measure >= p.FromMeas AND g.Measure <= p.ToMeas)
AND g.ReachCode=? AND g.Measure=?""", (reachcode, measure))
    result = cursor.fetchone()
    if None != result:
        comID = result[0]

    return comID


def getPlusFlowPredecessors(conn, comID):
    """ Get the immediate predecessors of the NHDPlus2 PlusFlow feature of comID
    
        @param conn A connection an SQLite3 database
        @param comdID String representing the ComID of the reach whose immediate predecessor reaches are to be discovered
        
        @return A list of immediate predecessor nodes in the NHDPlus2 PlusFlow graph
    """
    immediatePredecessors = []
    cursor = conn.cursor()
    cursor.execute("""SELECT FROMCOMID FROM PlusFlow WHERE TOCOMID=?""", (comID,))
    for row in cursor:
        immediatePredecessors.append(row[0])
    return immediatePredecessors


def getUpstreamReachesSQL(conn, comID, allUpstreamReaches):
    """ Recursively searches PlusFlow table in an SQLite database for all stream reaches
        upstream of a given reach.
    
        @note This method has no return value. Upstream reaches discovered are appended to allUpstreamReaches list.
    
        @param conn A connection to an SQLite3 database
        @param comID The ComID of the reach whose upstream reaches are to be discovered
        @param allUpstreamReaches A list containing integers representing comIDs of upstream reaches
    """
    upstream_reaches = getPlusFlowPredecessors(conn, comID)
    if len(upstream_reaches) == 0:
        return
    
    if len(upstream_reaches) == 1 and upstream_reaches[0] == 0:
        # We're at a headwater reach
        return

    # Foreach reach upstream of this reach
    for u in upstream_reaches:
        # Record the upstream reach
        allUpstreamReaches.append(u)
        # Find reaches upstream of it
        getUpstreamReachesSQL(conn, u, allUpstreamReaches)


def getFirstOrderUpstreamReachesNotInSet(config, comID, comIdsInSet, maxdepth=30):
    """ Search for upstream reaches downstream of reaches in the specified set.
    
        @param config A Python ConfigParser containing the following
        sections and options:
            'NHDPLUS2' and option 'PATH_OF_NHDPLUS2_DB' (absolute path to
            SQLite3 DB of NHDFlow data)
        @param comID The ComID of the reach whose upstream reaches downstream of those in the set
        are to be discovered
        @param comIdsInSet A set containing candidate comids
        @param upstreamReaches List containing integers representing comIDs of upstream reaches in set comIdsInSet
        @param maxdepth Integer representing maximum depth of recursion
        
        @return Set containing first order upstream reaches in set
        
        @raise ConfigParser.NoSectionError
        @raise ConfigParser.NoOptionError
    """
    nhddbPath = config.get('NHDPLUS2', 'PATH_OF_NHDPLUS2_DB')
    if not os.access(nhddbPath, os.R_OK):
        raise IOError(errno.EACCES, "The database at %s is not readable" %
                      nhddbPath)
    nhddbPath = os.path.abspath(nhddbPath)
    
    # Connect to DB
    conn = sqlite3.connect(nhddbPath)
    
    upstreamReaches = set()
    depth = 0
    getFirstOrderUpstreamReachesNotInSetSQL(conn, comID, comIdsInSet, upstreamReaches, depth, maxdepth)
    return list(upstreamReaches)


def getFirstOrderUpstreamReachesNotInSetSQL(conn, comID, comIdsInSet, upstreamReaches, depth, maxdepth):
    """ Recursively search for upstream reaches downstream of reaches in the specified set.
    
        @param conn An sqlite3 connection to a database that has NHDPlus2 tables
        @param comID The ComID of the reach whose upstream reaches downstream of those in the set
        are to be discovered
        @param comIdsInSet A set containing candidate comids
        @param upstreamReaches Set containing integers representing comIDs of upstream reaches in set comIdsInSet
        @param depth Integer current depth
        @param maxdepth Integer representing maximum depth of recursion
        
    """
    if depth > maxdepth:
        return
    
    upstream_reaches = getPlusFlowPredecessors(conn, comID)
    if len(upstream_reaches) == 0:
        return
    
    if len(upstream_reaches) == 1 and upstream_reaches[0] == 0:
        # We're at a headwater reach
        return

    # Foreach reach upstream of this reach
    for u in upstream_reaches:
        #print("\timmediate upstream: %s" % (type(u),) )
        # Return if upstream reach is in set
        if u in comIdsInSet:
            return
        else:
            # Keep looking in other upstream branches for first order reaches in set
            upstreamReaches.add(u)
            getFirstOrderUpstreamReachesNotInSetSQL(conn, u, comIdsInSet, upstreamReaches, depth + 1, maxdepth)



def getFirstOrderUpstreamReachesInSet(config, comID, comIdsInSet, maxdepth=30):
    """ Search for first-order upstream reaches in the specified set.
    
        @param config A Python ConfigParser containing the following
        sections and options:
            'NHDPLUS2' and option 'PATH_OF_NHDPLUS2_DB' (absolute path to
            SQLite3 DB of NHDFlow data)
        @param comID The ComID of the reach whose first-order upstream reaches are to be discovered
        @param comIdsInSet A set containing candidate comids
        @param upstreamReaches List containing integers representing comIDs of upstream reaches in set comIdsInSet
        @param maxdepth Integer representing maximum depth of recursion
        
        @return Set containing first order upstream reaches in set
        
        @raise ConfigParser.NoSectionError
        @raise ConfigParser.NoOptionError
    """
    nhddbPath = config.get('NHDPLUS2', 'PATH_OF_NHDPLUS2_DB')
    if not os.access(nhddbPath, os.R_OK):
        raise IOError(errno.EACCES, "The database at %s is not readable" %
                      nhddbPath)
    nhddbPath = os.path.abspath(nhddbPath)
    
    # Connect to DB
    conn = sqlite3.connect(nhddbPath)
    
    upstreamReaches = set()
    depth = 0
    getFirstOrderUpstreamReachesInSetSQL(conn, comID, comIdsInSet, upstreamReaches, depth, maxdepth)
    return list(upstreamReaches)


def getFirstOrderUpstreamReachesInSetSQL(conn, comID, comIdsInSet, upstreamReaches, depth, maxdepth):
    """ Recursively search for first-order upstream reaches in the specified set.
    
        @param conn An sqlite3 connection to a database that has NHDPlus2 tables
        @param comID The ComID of the reach whose first-order upstream reaches are to be discovered
        @param comIdsInSet A set containing candidate comids
        @param upstreamReaches Set containing integers representing comIDs of upstream reaches in set comIdsInSet
        @param depth Integer current depth
        @param maxdepth Integer representing maximum depth of recursion
        
    """
    if depth > maxdepth:
        return
    
    upstream_reaches = getPlusFlowPredecessors(conn, comID)
    if len(upstream_reaches) == 0:
        return
    
    if len(upstream_reaches) == 1 and upstream_reaches[0] == 0:
        # We're at a headwater reach
        return

    # Foreach reach upstream of this reach
    for u in upstream_reaches:
        #print("\timmediate upstream: %s" % (type(u),) )
        # Record the upstream reach if it is in set
        if u in comIdsInSet:
            upstreamReaches.add(u)
        else:
            # Keep looking in other upstream branches for first order reaches in set
            getFirstOrderUpstreamReachesInSetSQL(conn, u, comIdsInSet, upstreamReaches, depth + 1, maxdepth)

        
def getBoundingBoxForCatchmentsForGage(config, outputDir, reachcode, measure, deleteIntermediateFiles=True):
    """ Get bounding box coordinates (in WGS 84) for the drainage area associated with a given NHD 
        (National Hydrography Dataset) streamflow gage identified by a reach code and measure.
        
        @param config A Python ConfigParser containing the following sections and options:
            'GDAL/OGR' and option 'PATH_OF_OGR2OGR' (absolute path of ogr2ogr binary)
            'NHDPLUS2' and option 'PATH_OF_NHDPLUS2_DB' (absolute path to SQLite3 DB of NHDFlow data)
            'NHDPLUS2', 'PATH_OF_NHDPLUS2_CATCHMENT' (absolute path to NHD catchment SQLite3 spatial DB)
        @param outputDir String representing the absolute/relative path of the directory into which output rasters should be written
        @param reachcode String representing NHD streamflow gage 
        @param measure Float representing the measure along reach where Stream Gage is located 
            in percent from downstream end of the one or more NHDFlowline features that are 
            assigned to the ReachCode (see NHDPlusV21 GageLoc table)
        @param deleteIntermediateFiles A boolean, True if intermediate files generated from the analysis should be deleted
         
        @return A dictionary with keys: minX, minY, maxX, maxY, srs. The key srs is set to 'EPSG:4326' (WGS 84)
        
        @raise ConfigParser.NoSectionError
        @raise ConfigParser.NoOptionError
        @raise IOError(errno.ENOTDIR) if outputDir is not a directory
        @raise IOError(errno.EACCESS) if outputDir is not writable
    """
    nhddbPath = config.get('NHDPLUS2', 'PATH_OF_NHDPLUS2_DB')
    if not os.access(nhddbPath, os.R_OK):
        raise IOError(errno.EACCES, "The database at %s is not readable" %
                      nhddbPath)
    nhddbPath = os.path.abspath(nhddbPath)
        
    catchmentFeatureDBPath = config.get('NHDPLUS2', 'PATH_OF_NHDPLUS2_CATCHMENT')
    if not os.access(catchmentFeatureDBPath, os.R_OK):
        raise IOError(errno.EACCES, "The catchment feature DB at %s is not readable" %
                      catchmentFeatureDBPath)
    catchmentFeatureDBPath = os.path.abspath(catchmentFeatureDBPath)
    
    ogrCmdPath = config.get('GDAL/OGR', 'PATH_OF_OGR2OGR')
    if not os.access(ogrCmdPath, os.X_OK):
        raise IOError(errno.EACCES, "The ogr2ogr binary at %s is not executable" %
                      ogrCmdPath)
    ogrCmdPath = os.path.abspath(ogrCmdPath)
    
    if not os.path.isdir(outputDir):
        raise IOError(errno.ENOTDIR, "Output directory %s is not a directory" % (outputDir,))
    if not os.access(outputDir, os.W_OK):
        raise IOError(errno.EACCES, "Not allowed to write to output directory %s" % (outputDir,))
    outputDir = os.path.abspath(outputDir)
    
    # Connect to DB
    conn = sqlite3.connect(nhddbPath)
    
    comID = getComIdForStreamGage(conn, reachcode, measure)
    #sys.stderr.write("Gage with reachcode %s, measure %f has ComID %d" % (reachcode, measure, comID))
    
    # Get upstream reaches
    upstream_reaches = []
    getUpstreamReachesSQL(conn, comID, upstream_reaches)
    #sys.stderr.write("Upstream reaches: ")
    #sys.stderr.write(upstream_reaches)
    
    # Extract polygons for upstream catchments
    catchmentOut = os.path.join(outputDir, "catchment-%s.shp" % time.time())
    ogrCommand = "%s -s_srs EPSG:4326 -t_srs EPSG:4326  -f 'ESRI Shapefile' -sql 'SELECT * FROM catchment WHERE featureid=%s" % (ogrCmdPath, comID) # NHDPlusV2
    for reach in upstream_reaches:
        ogrCommand = ogrCommand + " OR featureid=%s" % reach # NHDPlusV2
    ogrCommand = ogrCommand +"' " + catchmentOut + " " + catchmentFeatureDBPath  
    #sys.stderr.write("ogr command: %s" % ogrCommand)
    os.system(ogrCommand)
    
    bbox = getBoundingBoxForShapefile(catchmentOut)
    
    # Clean-up temporary shapefile/ancillary files
    if deleteIntermediateFiles:
        deleteShapefile(catchmentOut)
    
    conn.close()

    return bbox


def getCatchmentFeaturesForReaches(config, outputDir,
                                   catchmentFilename, reaches,
                                   format=OGR_SHAPEFILE_DRIVER_NAME):
    """ Get features (in WGS 84) for the drainage area associated with a
        set of NHD (National Hydrography Dataset) stream reaches.
        
        @param config A Python ConfigParser containing the following
        sections and options:
            'PATH_OF_NHDPLUS2_CATCHMENT' (absolute path to
            NHD catchment shapefile)
        @param outputDir String representing the absolute/relative
        path of the directory into which output rasters should be
        written
        @param catchmentFilename String representing name of file to
        save catchment features to.  The appropriate extension will be added to the file name
        @param reaches List representing catchment features to be output
        @param format String representing OGR driver to use
        
        @return String representing the name of the dataset in outputDir created to hold
        the features
         
        @raise ConfigParser.NoSectionError
        @raise ConfigParser.NoOptionError
        @raise IOError(errno.ENOTDIR) if outputDir is not a directory
        @raise IOError(errno.EACCESS) if outputDir is not writable
        @raise Exception if output format is not known
        
        @todo Detect and fix non-closed geometries, e.g.
        kalisti:archive miles$ ./GetCatchmentsForComidsSP.py -p test -c 10462287
        Traceback (most recent call last):
          File "./GetCatchmentsForComidsSP.py", line 29, in <module>
            catchmentFilename, comid)
          File "/Users/miles/Dropbox/EarthCube-Multilayered/RHESSys-workflow/eclipse/EcohydroWorkflowLib/ecohydrolib/nhdplus2/networkanalysis.py", line 506, in getCatchmentFeaturesForComid
            outGeom = outGeom.Union( inGeom )
          File "/usr/local/Cellar/python/2.7.5/Frameworks/Python.framework/Versions/2.7/lib/python2.7/site-packages/osgeo/ogr.py", line 4065, in Union
            return _ogr.Geometry_Union(self, *args)
        RuntimeError: TopologyException: found non-noded intersection between LINESTRING (-77.9145 37.0768, -77.9147 37.0768) and LINESTRING (-77.9147 37.0768, -77.9145 37.0768) at -77.914621661942761 37.076822779115943
    
    """
    catchmentFeatureDBPath = config.get('NHDPLUS2', 'PATH_OF_NHDPLUS2_CATCHMENT')
    if not os.access(catchmentFeatureDBPath, os.R_OK):
        raise IOError(errno.EACCES, "The catchment feature DB at %s is not readable" %
                      catchmentFeatureDBPath)
    catchmentFeatureDBPath = os.path.abspath(catchmentFeatureDBPath)
    
    if not os.path.isdir(outputDir):
        raise IOError(errno.ENOTDIR, "Output directory %s is not a directory" % (outputDir,))
    if not os.access(outputDir, os.W_OK):
        raise IOError(errno.EACCES, "Not allowed to write to output directory %s" % (outputDir,))
    outputDir = os.path.abspath(outputDir)
    
    if not format in OGR_DRIVERS.keys():
        raise Exception("Output format '%s' is not known" % (format,) )
    
    catchmentFilename ="%s%s%s" % ( catchmentFilename, os.extsep, OGR_DRIVERS[format] )
    catchmentFilepath = os.path.join(outputDir, catchmentFilename)
    
    # Open input layer
    ogr.UseExceptions()
    poDS = ogr.Open(catchmentFeatureDBPath, OGR_UPDATE_MODE)
    if not poDS:
        raise Exception("Unable to open catchment feature database %s" (catchmentFeatureDBPath,))
    assert(poDS.GetLayerCount() > 0)
    poLayer = poDS.GetLayer(0)
    assert(poLayer)
    
    # Create output data source
    poDriver = ogr.GetDriverByName(format)
    assert(poDriver)
    poODS = poDriver.CreateDataSource(catchmentFilepath)
    assert(poODS != None)
#    poOLayer = poODS.CreateLayer("catchment", poLayer.GetSpatialRef(), poLayer.GetGeomType())
    poOLayer = poODS.CreateLayer("catchment", poLayer.GetSpatialRef(), ogr.wkbMultiPolygon )
#    poOLayer = poODS.CreateLayer("catchment", poLayer.GetSpatialRef(), ogr.wkbPolygon )
    
    # Create fields in output layer
    layerDefn = poLayer.GetLayerDefn()
    i = 0
    fieldCount = layerDefn.GetFieldCount()
    while i < fieldCount:
        fieldDefn = layerDefn.GetFieldDefn(i)
        poOLayer.CreateField(fieldDefn)
        i = i + 1

    # Create single geometry to hold catchment polygon in output shapefile
    outGeom = ogr.Geometry( poOLayer.GetGeomType() )
#    polygon = Polygon()

    # Copy features, unioning them as we go
    numReaches = len(reaches)
    # Copy features in batches of UPSTREAM_SEARCH_THRESHOLD to overcome limit in 
    #   OGR driver for input layer
    start = 0
    end = UPSTREAM_SEARCH_THRESHOLD
    while end < numReaches:
        whereFilter = "featureid=%s" % (reaches[start],)
        for reach in reaches[start+1:end]:
            whereFilter = whereFilter + " OR featureid=%s" % (reach,)
        # Copy features
        assert(poLayer.SetAttributeFilter(whereFilter) == 0)
        inFeature = poLayer.GetNextFeature()
        # Union geometry of input feature to output feature
        while inFeature:
#            inGeom = inFeature.GetGeometryRef().SimplifyPreserveTopology(0.0001)
            inGeom = inFeature.GetGeometryRef()
            outGeom = outGeom.Union( inGeom )
#            polygon = polygon.union( loads( inGeom.ExportToWkb() ) )
#            polygon = cascaded_union( [polygon, loads( inGeom.ExportToWkb() )] )
            inFeature.Destroy()
            inFeature = poLayer.GetNextFeature() 
        start = end
        end = end + UPSTREAM_SEARCH_THRESHOLD
    # Copy remaining features
    whereFilter = "featureid=%s" % (reaches[start],)
    for reach in reaches[start+1:end]:
        whereFilter = whereFilter + " OR featureid=%s" % (reach,)
    # Copy features
    poLayer.SetAttributeFilter(whereFilter)
    assert(poLayer.SetAttributeFilter(whereFilter) == 0)
    inFeature = poLayer.GetNextFeature()
    while inFeature:
#        inGeom = inFeature.GetGeometryRef().SimplifyPreserveTopology(0.0001)
        inGeom = inFeature.GetGeometryRef()
        outGeom = outGeom.Union( inGeom )
#        polygon = polygon.union( loads( inGeom.ExportToWkb() ) )
#        polygon = cascaded_union( [polygon, loads( inGeom.ExportToWkb() )] )
        inFeature.Destroy()
        inFeature = poLayer.GetNextFeature()
    
    # Create a new polygon that only contains exterior points
    outGeom = ogr.ForceToPolygon( outGeom )
    polygon = loads( outGeom.ExportToWkb() )
    if polygon.exterior:
        coords = polygon.exterior.coords
        newPolygon = Polygon(coords)
    else:
        newPolygon = Polygon()
    
    # Write new feature to output feature data source
    outFeat = ogr.Feature( poOLayer.GetLayerDefn() )
    outFeat.SetGeometry( ogr.CreateGeometryFromWkb( dumps(newPolygon) ) )
    poOLayer.CreateFeature(outFeat)
        
    return catchmentFilename


def getCatchmentFeaturesForComid(config, outputDir,
                                catchmentFilename, comID,
                                format=OGR_SHAPEFILE_DRIVER_NAME):
    """ Get features (in WGS 84) for the drainage area associated with a
        given NHD (National Hydrography Dataset) stream reach.
         
        @param config A Python ConfigParser containing the following
        sections and options:
            'NHDPLUS2',
        @param outputDir String representing the absolute/relative
        path of the directory into which output rasters should be
        written
        @param catchmentFilename String representing name of file to
        save catchment features to.  The appropriate extension will be added to the file name
        @param comID String representing comid of stream reach whose upstream
        catchment area is to be determined
        @param format String representing OGR driver to use
        
        @return String representing the name of the dataset in outputDir created to hold
        the features
         
        @raise ConfigParser.NoSectionError
        @raise ConfigParser.NoOptionError
        @raise IOError(errno.ENOTDIR) if outputDir is not a directory
        @raise IOError(errno.EACCESS) if outputDir is not writable
        @raise Exception if output format is not known
        
    """
    nhddbPath = config.get('NHDPLUS2', 'PATH_OF_NHDPLUS2_DB')
    if not os.access(nhddbPath, os.R_OK):
        raise IOError(errno.EACCES, "The database at %s is not readable" %
                      nhddbPath)
    nhddbPath = os.path.abspath(nhddbPath)
    
    # Connect to DB
    conn = sqlite3.connect(nhddbPath)
    
    # Get upstream reaches
    reaches = [comID]
    getUpstreamReachesSQL(conn, comID, reaches)
    #sys.stderr.write("Upstream reaches: ")
    #sys.stderr.write(upstream_reaches)
    conn.close()
    
    return getCatchmentFeaturesForReaches(config, outputDir,
                                   catchmentFilename, reaches,
                                   format)

 
def getCatchmentFeaturesForGage(config, outputDir,
                                catchmentFilename, reachcode, measure, 
                                format=OGR_SHAPEFILE_DRIVER_NAME):
    """ Get features (in WGS 84) for the drainage area associated with a
        given NHD (National Hydrography Dataset) streamflow gage
        identified by a reach code and measure.
        
        @note Capable of handling gages with more than 1000 upstream reaches 
        
        @note No return value. catchmentFilename will be written to
        outputDir if successful
        
        @param config A Python ConfigParser containing the following
        sections and options:
            'NHDPLUS2' and option 'PATH_OF_NHDPLUS2_DB' (absolute path to
            SQLite3 DB of NHDFlow data)

            'NHDPLUS2', 'PATH_OF_NHDPLUS2_CATCHMENT' (absolute path to
            NHD catchment shapefile)
        @param outputDir String representing the absolute/relative
        path of the directory into which output rasters should be
        written
        @param catchmentFilename String representing name of file to
        save catchment features to.  The appropriate extension will be added to the file name
        @param reachcode String representing NHD streamflow gage 
        @param measure Float representing the measure along reach
        where Stream Gage is located in percent from downstream
        end of the one or more NHDFlowline features that are
        assigned to the ReachCode (see NHDPlusV21 GageLoc table)
        @param format String representing OGR driver to use
        
        @return String representing the name of the dataset in outputDir created to hold
        the features
         
        @raise ConfigParser.NoSectionError
        @raise ConfigParser.NoOptionError
        @raise IOError(errno.ENOTDIR) if outputDir is not a directory
        @raise IOError(errno.EACCESS) if outputDir is not writable
        @raise Exception if output format is not known
    """
    nhddbPath = config.get('NHDPLUS2', 'PATH_OF_NHDPLUS2_DB')
    if not os.access(nhddbPath, os.R_OK):
        raise IOError(errno.EACCES, "The database at %s is not readable" %
                      nhddbPath)
    nhddbPath = os.path.abspath(nhddbPath)
    
    # Connect to DB
    conn = sqlite3.connect(nhddbPath)
    
    comID = getComIdForStreamGage(conn, reachcode, measure)
    #sys.stderr.write("Gage with reachcode %s, measure %f has ComID %d" % (reachcode, measure, comID))
    
    return getCatchmentFeaturesForComid(config, outputDir,
                                catchmentFilename, comID,
                                format)



"""@package ecohydrolib.climatedata.ghcndquery
    
@brief Query NCDC Global Historical Climatology Network dataset for daily
climate data

@note Requires pyspatialite 3.0.1

This software is provided free of charge under the New BSD License. Please see
the following license information:

Copyright (c) 2013, University of North Carolina at Chapel Hill
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of the University of North Carolina at Chapel Hill nor the
      names of its contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE UNIVERSITY OF NORTH CAROLINA AT CHAPEL HILL
BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

@author Brian Miles <brian_miles@unc.edu>
"""
import os, errno
import re
import httplib
from pyspatialite import dbapi2 as spatialite


# Example URL http://www1.ncdc.noaa.gov/pub/data/ghcn/daily/all/US1NCDH0006.dly
HOST = 'www1.ncdc.noaa.gov'
URL_PROTO = '/pub/data/ghcn/daily/all/{station_id}.dly'

_SRS = int(4326)
_BUFF_LEN = 4096 * 10


def findStationsWithinBoundingBox(config, bbox):
    """ Find stations that lie within a bounding box
    
        @param config ConfigParser containing the section 'GHCND' and option 
        'PATH_OF_STATION_DB'
        @param bbox A dict containing keys: minX, minY, maxX, maxY, srs, where srs='EPSG:4326'
        
        @return A list of GHCN station attributes for each station within the bounding box:
        [id, lat, lon, elevation, name]
        
        @code
        import os
        import ConfigParser
        from ecohydrolib.metadata import GenericMetadata
        from ecohydrolib.climatedata.ghcndquery import findStationsWithinBoundingBox
        studyArea = GenericMetadata.readStudyAreaEntries(projectDir)
        bbox = studyArea['bbox_wgs84'].split()
        bbox = dict({'minX': float(bbox[0]), 'minY': float(bbox[1]), 'maxX': float(bbox[2]), 'maxY': float(bbox[3]), 'srs': 'EPSG:4326'})
            
        configFile = os.environ['ECOHYDROWORKFLOW_CFG']
        config = ConfigParser.RawConfigParser()
        config.read(configFile)
        
        stations = findStationsWithinBoundingBox(config,bbox)
        @endcode 
    """
    stations = []
    pattern = re.compile("^POINT\((-?\d+\.?\d*) (-?\d+\.?\d*)\)$")
    
    ghcnDB = config.get('GHCND', 'PATH_OF_STATION_DB')
    
    conn = spatialite.connect(ghcnDB)
    cursor = conn.cursor()
    # Spatialite/SQLite3 won't subsitute parameter strings within quotes, so we have to do it the unsafe way.  This should be okay as
    # we are dealing with numeric values that we are converting to numeric types before building the query string.
    sql = u"SELECT id,AsText(coord),elevation_m,name FROM ghcn_station WHERE Within(coord, BuildMbr(%f,%f,%f,%f));" %\
    (bbox['minX'], bbox['minY'], bbox['maxX'], bbox['maxY'])
    cursor.execute(sql)
    results = cursor.fetchall()
    conn.close()
    for result in results:
        # Unpack the coordinates
        match = pattern.match(result[1])
        assert(match)
        lon = float(match.group(1))
        lat = float(match.group(2))
        stations.append([result[0], lat, lon, result[2], result[3]])
    return stations


def findStationNearestToCoordinates(config, longitude, latitude):
    """Determine identifier of station nearest to longitude, latitude coordinates.
    
        @param config ConfigParser containing the section 'GHCND' and option 
        'PATH_OF_STATION_DB'
        @param longitude Float representing WGS84 longitude
        @param latitude Float representing WGS84 latitude
        
        @return Tuple of the form (station_id, longitude, latitude, elevation_meters, name, distance),
        None if no gage is found.
        
        @code
        from ecohydrolib.climatedata.ghcndquery import findStationNearestToCoordinates
        import ConfigParser
        config = ConfigParser.RawConfigParser()
        config.read('./bin/macosx2.cfg')
        lon = -76.7443397486
        lat = 39.2955590994
        nearest = findStationNearestToCoordinates(config, lon, lat)
        from ecohydrolib.climatedata.ghcndquery import getClimateDataForStation
        outputDir = '/tmp'
        outfileName = 'clim.txt'
        getClimateDataForStation(config, outputDir, outfileName, nearest[0])
        @endcode
    """
    ghcnDB = config.get('GHCND', 'PATH_OF_STATION_DB')
    
    conn = spatialite.connect(ghcnDB)
    cursor = conn.cursor()
    # Spatialite/SQLite3 won't subsitute parameter strings within quotes, so we have to do it the unsafe way.  This should be okay as
    # we are dealing with numeric values that we are converting to numeric types before building the query string.
    sql = u"SELECT id,AsText(coord),elevation_m,name,Distance(GeomFromText('POINT(%f %f)', %d), coord) as dist FROM ghcn_station order by dist asc limit 1" %\
    (float(longitude), float(latitude), _SRS)
    cursor.execute(sql)
    nearest = cursor.fetchone()
    conn.close()
    if nearest:
        # Unpack the coordinates
        pattern = re.compile("^POINT\((-?\d+\.\d+) (-?\d+\.\d+)\)$")
        match = pattern.match(nearest[1])
        assert(match)
        lon = float(match.group(1))
        lat = float(match.group(2))
        return (nearest[0], lon, lat, nearest[2], nearest[3], nearest[4])
    return None
    

def getClimateDataForStation(config, outputDir, outFilename, stationID, overwrite=True):
    """Fetch climate timeseries data for a GHCN daily station
    
        @param config A Python ConfigParser (not currently used)
        @param outputDir String representing the absolute/relative path of the directory into which output DEM should be written
        @param outDEMFilename String representing the name of the DEM file to be written
        @param stationID String representing unique identifier of station
        @param overwrite Boolean value indicating whether or not the file indicated by filename should be overwritten.
            If False and filename exists, IOError exception will be thrown with errno.EEXIST
        @param overwrite Boolean value indicating whether or not the file indicated by filename should be overwritten.
            If False and filename exists, IOError exception will be thrown with errno.EEXIST
    
        @raise IOError if outputDir is not a writable directory
        @raise IOError if outFilename already exists and overwrite is False (see above)
        
        @return True if timeseries data were fetched and False if not
    """
    dataFetched = False
    
    if not os.path.isdir(outputDir):
        raise IOError(errno.ENOTDIR, "Output directory %s is not a directory" % (outputDir,))
    if not os.access(outputDir, os.W_OK):
        raise IOError(errno.EACCES, "Not allowed to write to output directory %s" % (outputDir,))
    outputDir = os.path.abspath(outputDir)
    
    outFilepath = os.path.join(outputDir, outFilename)
    
    if os.path.exists(outFilepath):
        if overwrite: 
            os.unlink(outFilepath)
        else:
            raise IOError(errno.EEXIST, "File %s already exists" % outFilepath)
    
    url = URL_PROTO.format(station_id=stationID)
    
    conn = httplib.HTTPConnection(HOST)
    conn.request('GET', url)
    res = conn.getresponse(buffering=True)
    
    assert(200 == res.status)
    
    data = res.read(_BUFF_LEN)
    if data: 
        dataOut = open(outFilepath, 'wb')
        dataFetched = True
        while data:
            dataOut.write(data)
            data = res.read(_BUFF_LEN)
        dataOut.close()
        
    return dataFetched


"""@package ecohydrolib.tests.test_nhdwebsvc
    
    @brief Test methods for ecohydrolib.nhdplus2.webservice
    
    This software is provided free of charge under the New BSD License. Please see
    the following license information:
    
    Copyright (c) 2013, University of North Carolina at Chapel Hill
    All rights reserved.
    
    Redistribution and use in source and binary forms, with or without
    modification, are permitted provided that the following conditions are met:
        * Redistributions of source code must retain the above copyright
          notice, this list of conditions and the following disclaimer.
        * Redistributions in binary form must reproduce the above copyright
          notice, this list of conditions and the following disclaimer in the
          documentation and/or other materials provided with the distribution.
        * Neither the name of the University of North Carolina at Chapel Hill nor the
          names of its contributors may be used to endorse or promote products
          derived from this software without specific prior written permission.
    
    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
    ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
    WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
    DISCLAIMED. IN NO EVENT SHALL THE UNIVERSITY OF NORTH CAROLINA AT CHAPEL HILL
    BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
    CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
    GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
    HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
    LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
    OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


    @author Brian Miles <brian_miles@unc.edu>
    
    Usage: 
    @code
    python -m unittest test_metadata
    @endcode
    
""" 
import unittest
import traceback
import os, sys
import tempfile, shutil

from ecohydrolib.context import Context

from ecohydrolib.nhdplus2.webservice import WebserviceError
from ecohydrolib.nhdplus2.webservice import getCatchmentFeaturesForStreamflowGage
from ecohydrolib.nhdplus2.webservice import locateStreamflowGage
from ecohydrolib.nhdplus2.webservice import RESPONSE_OK

class Test(unittest.TestCase):

    def setUp(self):
        self.projectDir = tempfile.mkdtemp()
        self.context = Context(projectDir=self.projectDir)
        
    def tearDown(self):
        shutil.rmtree(self.projectDir)

    def testLocateStreamflowGage(self):
        badGage = "3ti04ti04tgi0"
        (response, url) = locateStreamflowGage(self.context.config, badGage)
        self.assertTrue( response['message'] == "Illegal gageid '%s'" % (badGage,) )
        
        unknownGage = "123456789101112"
        (response, url) = locateStreamflowGage(self.context.config, unknownGage)
        self.assertTrue( response['message'] == "Reachcode for gage '%s' not found" % (unknownGage,) )

        goodGage = "01589312"
        (response, url) = locateStreamflowGage(self.context.config, goodGage)
        self.assertTrue( response['message'] == RESPONSE_OK )
        self.assertTrue( response['measure'] == 75.06239 )
        self.assertTrue( response['gage_lon'] == -76.74433974864309 )
        self.assertTrue( response['gage_lat'] == 39.295559099434264 )
        self.assertTrue( response['reachcode'] == "02060003000745" )

    def testGetCatchmentFeaturesForStreamflowGage(self):
        goodReachcode = "02060003000745"
        goodMeasure = 75.06239
        badReachcode = "klsdklasd"
        badMeasure = "sdfklfkl"

        # Bad reachcode
        exceptionThrown = False
        try:
            (shapefileName, url) = getCatchmentFeaturesForStreamflowGage(self.context.config, self.projectDir,
                                                                         'catchment', badReachcode, goodMeasure)
        except WebserviceError:
            exceptionThrown = True
        except Exception:
            raise
        
        self.assertTrue( exceptionThrown )
        
        # Bad measure
        exceptionThrown = False
        try:
            (shapefileName, url) = getCatchmentFeaturesForStreamflowGage(self.context.config, self.projectDir,
                                                                         'catchment', goodReachcode, badMeasure)
        except WebserviceError:
            exceptionThrown = True
        except Exception:
            raise
        
        self.assertTrue( exceptionThrown )
        
        # Good case
        exceptionThrown = False
        try:
            (shapefileName, url) = getCatchmentFeaturesForStreamflowGage(self.context.config, self.projectDir,
                                                                         'catchment', goodReachcode, goodMeasure)
        except:
            traceback.print_exc(file=sys.stdout)
            exceptionThrown = True
        
        self.assertTrue( not exceptionThrown )
        self.assertTrue( shapefileName == 'catchment.shp' )
        self.assertTrue( os.path.exists( os.path.join( self.projectDir, shapefileName ) ) )

if __name__ == "__main__":
    #import sys;sys.argv = ['', 'Test.testName']
    unittest.main()


"""@package ecohydrolib.wcs4dem.demquery
    
@brief Query NASA EOS Education Alliance (NEHEA) GeoBrain for DEM data

This software is provided free of charge under the New BSD License. Please see
the following license information:

Copyright (c) 2013, University of North Carolina at Chapel Hill
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of the University of North Carolina at Chapel Hill nor the
      names of its contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE UNIVERSITY OF NORTH CAROLINA AT CHAPEL HILL
BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


@author Brian Miles <brian_miles@unc.edu>
"""
from ecohydrolib.wcslib import getRasterForBoundingBox

COVERAGE_NED_30m_USA = 'NED_30m_USA'
COVERAGE_SRTM_30m_USA = 'SRTM_30m_USA'
COVERAGE_SRTM_90m_USA = 'SRTM_90m_Global'
COVERAGE_GTOPO_30arc_GLOBAL = 'GTOPO_30arc_Global'
SUPPORTED_COVERAGE = [COVERAGE_NED_30m_USA, COVERAGE_SRTM_30m_USA, COVERAGE_SRTM_90m_USA, COVERAGE_GTOPO_30arc_GLOBAL]

FORMAT_GEOTIFF = 'image/geotiff'
FORMAT_NETCDF = 'image/netcdf'
FORMAT_PNG = 'image/PNG'
FORMAT_JPEG = 'image/JPEG'
FORMAT_JP2000 = 'image/JPEG2000'
FORMAT_HDF4 = 'image/HDF4IMAGE'
SUPPORTED_FORMATS = [FORMAT_GEOTIFF, FORMAT_NETCDF, FORMAT_PNG, FORMAT_JPEG, FORMAT_JP2000, FORMAT_HDF4]

MIME_TYPE = 'application/image'

# Example URL /cgi-bin/gbwcs-dem?service=wcs&version=1.0.0&request=getcoverage&coverage=SRTM_90m_Global&bbox=-90,38,-89,39&crs=epsg:4326&format=image/geotiff&store=true
HOST = 'geobrain.laits.gmu.edu'
URL_PROTO = '/cgi-bin/gbwcs-dem?service=wcs&version=1.0.0&request=getcoverage&coverage={coverage}&crs={crs}&bbox={bbox}&response_crs={response_crs}&format={format}&store={store}'

def getDEMForBoundingBox(config, outputDir, outDEMFilename, bbox, coverage=COVERAGE_NED_30m_USA, srs='EPSG:4326', format=FORMAT_GEOTIFF, overwrite=True):
    """ Fetch a digital elevation model (DEM) from the GeoBrain WCS4DEM WCS-compliant web service.
        Will write any error returned by query to sys.stderr.
    
        @param config A Python ConfigParser (not currently used)
        @param outputDir String representing the absolute/relative path of the directory into which output DEM should be written
        @param outDEMFilename String representing the name of the DEM file to be written
        @param bbox Dict representing the lat/long coordinates and spatial reference of the bounding box area
            for which the DEM is to be extracted.  The following keys must be specified: minX, minY, maxX, maxY, srs.
        @param coverage String representing the DEM source from which to get the DEM coverage.  Must be a value listed in SUPPORTED_COVERAGE
        @param srs String representing the spatial reference of the raster to be returned.
        @param format String representing the MIME type of the raster format to be returned.  Must be a value listed in 
            SUPPORTED_FORMATS
        @param overwrite Boolean value indicating whether or not the file indicated by filename should be overwritten.
            If False and filename exists, IOError exception will be thrown with errno.EEXIST
    
        @raise IOError if outputDir is not a writable directory
        @raise IOError if outDEMFilename already exists and overwrite is False (see above)
    
        @return Tuple(True if DEM data were fetched and False if not, URL of DEM fetched)
    """
    assert(format in SUPPORTED_FORMATS)
    assert(coverage in SUPPORTED_COVERAGE)
    return getRasterForBoundingBox(config, outputDir, outDEMFilename, HOST, URL_PROTO, MIME_TYPE, bbox, coverage, srs, format,
                                   response_crs=srs, store=False, overwrite=overwrite)
    


"""@package ecohydrolib.context
    
@brief Class that encapsulated environment setup for ecohydrolib scripts

This software is provided free of charge under the New BSD License. Please see
the following license information:

Copyright (c) 2013, University of North Carolina at Chapel Hill
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of the University of North Carolina at Chapel Hill nor the
      names of its contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE UNIVERSITY OF NORTH CAROLINA AT CHAPEL HILL
BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


@author Brian Miles <brian_miles@unc.edu>
"""
import os, errno
import ConfigParser

from ecohydrolib.metadata import GenericMetadata

CONFIG_FILE_ENV = 'ECOHYDROLIB_CFG'

class Context(object):
    def __init__(self, projectDir, configFile=None):
        """ Constructor for Context class
        
            @param projectDir Path of the project whose metadata store is to be read from
            @param configFile Path of ecohydrolib configuration file to use.  If none,
            will attempt to read configuration from a file named in the environment
            variable Context.CONFIG_FILE_ENV
            
            @raise IOError if project directory path is not a directory
            @raise IOError if project directory is not writable
            @raise MetadataVersionError if a version already exists in the 
            metadata store and is different than GenericMetadata._ecohydrolibVersion
            @raise EnvironmentError if configuration file name could not be read from the 
            environment
            @raise IOError if configuration file could not be read
        """
        if not os.path.isdir(projectDir):
            raise IOError(errno.ENOTDIR, "Specified project directory %s is not a directory" % \
                          (projectDir,))
        if not os.access(projectDir, os.W_OK):
            raise IOError(errno.EACCES, "Unable to write to project directory %s" % \
                          (projectDir,))
        self.projectDir = os.path.abspath(projectDir)
        
        # Make sure metadata version is compatible with this version of ecohydrolib
        #   will raise MetadataVersionError if there is a version mismatch
        GenericMetadata.checkMetadataVersion(projectDir)
        
        if not configFile:
            try:
                self._configFile = os.environ[CONFIG_FILE_ENV]
            except KeyError:
                raise EnvironmentError("Configuration file not specified via environmental variable %s" %\
                                       CONFIG_FILE_ENV)
        else:
            self._configFile = configFile
        
        if not os.access(self._configFile, os.R_OK):
            raise IOError(errno.EACCES, "Unable to read configuration file %s" %
                          self._configFile)
        self.config = ConfigParser.RawConfigParser()
        self.config.read(self._configFile)


#!/usr/bin/env python
"""@package GetCatchmentShapefileForNHDStreamflowGage

@brief Query NHDPlus2 database for shapefile of the drainage area of the given streamflow gage.
@brief Resulting shapefile will use WGS84 (EPSG:4326) for its spatial reference. 

This software is provided free of charge under the New BSD License. Please see
the following license information:

Copyright (c) 2013, University of North Carolina at Chapel Hill
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of the University of North Carolina at Chapel Hill nor the
      names of its contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE UNIVERSITY OF NORTH CAROLINA AT CHAPEL HILL
BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


@author Brian Miles <brian_miles@unc.edu>


Pre conditions
--------------
1. Configuration file must define the following sections and values:
   'NHDPLUS2', 'PATH_OF_NHDPLUS2_DB'
   'NHDPLUS2', 'PATH_OF_NHDPLUS2_CATCHMENT'

2. The following metadata entry(ies) must be present in the study area section of the metadata associated with the project directory:
   nhd_gage_reachcode
   nhd_gage_measure_pct 

Post conditions
---------------
1. Will write the following entry(ies) to the manifest section of metadata associated with the project directory:
   study_area_shapefile [the name of the catchment shapefile] 

Usage:
@code
GetCatchmentShapefileForNHDStreamflowGage.py -p /path/to/project_dir
@endcode

@note EcohydroLib configuration file must be specified by environmental variable 'ECOHYDROWORKFLOW_CFG',
or -i option must be specified. 
"""
import os
import sys
import argparse
import textwrap

from ecohydrolib.context import Context
from ecohydrolib.metadata import GenericMetadata
from ecohydrolib.metadata import AssetProvenance

from ecohydrolib.spatialdata.utils import deleteShapefile

from ecohydrolib.nhdplus2.webservice import getCatchmentFeaturesForStreamflowGage
from ecohydrolib.nhdplus2.networkanalysis import getCatchmentFeaturesForGage
from ecohydrolib.nhdplus2.networkanalysis import OGR_DRIVERS
from ecohydrolib.nhdplus2.networkanalysis import OGR_SHAPEFILE_DRIVER_NAME

# Handle command line options
parser = argparse.ArgumentParser(description='Get shapefile for the drainage area of an NHDPlus2 streamflow gage')
parser.add_argument('-i', '--configfile', dest='configfile', required=False,
                    help='The configuration file')
parser.add_argument('-p', '--projectDir', dest='projectDir', required=True,
                    help='The directory to which metadata, intermediate, and final files should be saved')
parser.add_argument('-s', '--source', dest='source', required=False, choices=['local', 'webservice'], default='webservice',
                    help='Source to query NHDPlusV2 dataset')
parser.add_argument('-f', '--outfile', dest='outfile', required=False,
                    help='The name of the catchment shapefile to be written.  File extension ".shp" will be added. ' +
                    ' If a file of this name exists, program will silently exit.')
parser.add_argument('--overwrite', dest='overwrite', action='store_true', required=False,
                    help='Overwrite existing catchment shapefile in project directory.  If not specified, program will halt if a dataset already exists.')
args = parser.parse_args()
cmdline = GenericMetadata.getCommandLine()

configFile = None
if args.configfile:
    configFile = args.configfile

context = Context(args.projectDir, configFile) 

if args.outfile:
    outfile = args.outfile
else:
    outfile = "catchment" 

tmpFilename = "%s%s%s" % (outfile, os.extsep, OGR_DRIVERS[OGR_SHAPEFILE_DRIVER_NAME])
shapeFilepath = os.path.join(context.projectDir, tmpFilename)

if not args.overwrite:
    if os.path.exists(shapeFilepath):
        sys.exit( textwrap.fill("Catchment shapefile already exists in project directory %s.  Use --overwrite option to overwrite." % \
                 args.projectDir ) )
elif os.path.exists(shapeFilepath):
    # Overwrite was specified
    deleteShapefile(shapeFilepath)

# Get provenance data for gage
gageProvenance = [i for i in GenericMetadata.readAssetProvenanceObjects(context) if i.name == 'gage'][0]
if gageProvenance is None:
    sys.exit("Unable to load gage provenance information from metadata")

# Get study area parameters
studyArea = GenericMetadata.readStudyAreaEntries(context)
reachcode = studyArea['nhd_gage_reachcode']
measure = studyArea['nhd_gage_measure_pct']

writeMetadata = False
if args.source == 'local':
    sys.stdout.write('Getting catchment area draining through gage using local NHDPlus dataset...')
    sys.stdout.flush()
    if not context.config.has_option('NHDPLUS2', 'PATH_OF_NHDPLUS2_DB'):
        sys.exit("Config file %s does not define option %s in section %s" % \
              (args.configfile, 'NHDPLUS2', 'PATH_OF_NHDPLUS2_DB'))
    if not context.config.has_option('NHDPLUS2', 'PATH_OF_NHDPLUS2_CATCHMENT'):
        sys.exit("Config file %s does not define option %s in section %s" % \
              (args.configfile, 'NHDPLUS2', 'PATH_OF_NHDPLUS2_CATCHMENT')) 

    shapeFilename = getCatchmentFeaturesForGage(context.config, context.projectDir, outfile, 
                                                reachcode, measure,
                                                format=OGR_SHAPEFILE_DRIVER_NAME)
    source = 'http://www.horizon-systems.com/NHDPlus/NHDPlusV2_home.php'
    writeMetadata = True
    sys.stdout.write('done\n')
else:
    sys.stdout.write('Geting catchment area draining through gage using NHDPlus webservice...')
    sys.stdout.flush()
    
    try:
        (shapeFilename, source) = getCatchmentFeaturesForStreamflowGage(context.config, context.projectDir,
                                                                       outfile, reachcode, measure,
                                                                       format=OGR_SHAPEFILE_DRIVER_NAME)
        writeMetadata = True
    except Exception as e:
        sys.exit( str(e) )
    
    sys.stdout.write('done\n')

if writeMetadata:
    # Write provenance
    asset = AssetProvenance(GenericMetadata.MANIFEST_SECTION)
    asset.name = 'study_area_shapefile'
    asset.dcIdentifier = shapeFilename
    asset.dcSource = source
    asset.dcTitle = 'Study area shapefile'
    asset.dcPublisher = 'USGS'
    asset.dcDescription = cmdline
    asset.writeToMetadata(context)

    # Write processing history
    GenericMetadata.appendProcessingHistoryItem(context, cmdline)


"""@package ecohydrolib.solim.inference
    
@brief Infer soil properties from SSURGO and terrain data using SOLIM framework

This software is provided free of charge under the New BSD License. Please see
the following license information:

Copyright (c) 2013, University of North Carolina at Chapel Hill
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of the University of North Carolina at Chapel Hill nor the
      names of its contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE UNIVERSITY OF NORTH CAROLINA AT CHAPEL HILL
BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


@author Brian Miles <brian_miles@unc.edu>
"""
import os, errno


ATTRIBUTES = ['avgSand','avgSilt','avgClay','avgKsat','avgPorosity']
ATTRIBUTE_SEP = ','
FILE_EXT = 'tif'

def inferSoilPropertiesForSSURGOAndTerrainData(config, outputDir, shpFilepath, demFilepath, \
                                               featureAttrList=ATTRIBUTES):
    """ Infer soil properties from SSURGO and terrain data using SOLIM framework
    
        @param config ConfigParser containing the section 'SOLIM' and option 'PATH_OF_SOLIM'
        @param outputDir String representing the absolute/relative path of the directory into which shapefile should be written
        @param shpFilepath String representing the absolute path of the shapefile containing SSURGO features
        @param featureAttrList List containing the SSURGO attributes for which soil property inference is to be performed
        @param demFilepath String representing the absolute path of the DEM terrain data
        
        @return Dictionary containing the keys for each soil attribute and values of the names of the raster files generated for that attribute
        
        @exception IOError(errno.EACCES) if SOLIM binary is not executable
        @exception Exception if SOLIM command fails
    """
    solimCmdPath = config.get('SOLIM', 'PATH_OF_SOLIM')
    if not os.access(solimCmdPath, os.X_OK):
        raise IOError(errno.EACCES, "SOLIM command %s is not executable" % (solimCmdPath,))
    
    if not os.path.isdir(outputDir):
        raise IOError(errno.ENOTDIR, "Output directory %s is not a directory" % (outputDir,))
    if not os.access(outputDir, os.W_OK):
        raise IOError(errno.EACCES, "Not allowed to write to output directory %s" % (outputDir,))
    outputDir = os.path.abspath(outputDir)
    
    if not os.access(shpFilepath, os.R_OK):
        raise IOError(errno.EACCES, "Not allowed to read SSURGO feature shapefile %s" % (shpFilepath,))
    if not os.path.isfile(shpFilepath):
        raise IOError(errno.ENOENT, "SSURGO feature shapefile %s does not exist" % (shpFilepath,))
    shpFilepath = os.path.abspath(shpFilepath)
    
    if not os.access(demFilepath, os.R_OK):
        raise IOError(errno.EACCES, "Not allowed to read terrain DEM file %s" % (demFilepath,))
    if not os.path.isfile(demFilepath):
        raise IOError(errno.ENOENT, "Terrain DEM file %s does not exist" % (demFilepath,))
    demFilepath = os.path.abspath(demFilepath)
    
    solimCommand = "%s %s mukey %s %s %s" % \
        (solimCmdPath, shpFilepath, demFilepath, outputDir, ATTRIBUTE_SEP.join(featureAttrList))
    #print solimCommand
    returnCode = os.system(solimCommand)
    if returnCode != 0:
        raise Exception("SOLIM command %s failed, returning %d" % (solimCommand, returnCode) )
    
    filesCreated = dict()
    for attr in featureAttrList:
        filename = attr + os.extsep + FILE_EXT
        if not os.path.exists(os.path.join(outputDir, filename)):
            raise Exception("SOLIM failed to create file %s in directory %s" % \
                            (filename, outputDir) )
        filesCreated[attr] = filename
    
    return filesCreated